from enum import Enum


class ExampleTypeEnum(str, Enum):
    """Enum for different example project types available for initialization."""

    github_releases = "github-releases"
    ev_sales = "ev-sales"
    college_football = "college-football"



from visivo.discovery.discover import Discover
from visivo.parsers.parser_factory import ParserFactory
from visivo.models.project import Defaults, Project
from visivo.logger.logger import Logger
from visivo.commands.dbt_phase import dbt_phase
from visivo.utils import get_dashboards_dir
import yaml
import click
import os
from time import time


def parse_project_phase(
    working_dir, output_dir, default_source, dbt_profile, dbt_target, new=False, project_dir=""
):
    discover = Discover(working_dir=working_dir, output_dir=output_dir)
    project = None

    if not os.path.exists(discover.project_file) or new:
        """
        Ignore project parse if it's quickstart
        """
        project = Project(
            name="Quickstart Visivo",
            sources=[],
            models=[],
            traces=[],
            charts=[],
            dashboards=[],
        )
    else:
        # Run and Track dbt phase
        dbt_start = time()
        Logger.instance().debug("    Running dbt phase...")
        dbt_phase(working_dir, output_dir, dbt_profile, dbt_target)
        dbt_duration = round(time() - dbt_start, 2)
        if os.environ.get("STACKTRACE"):
            Logger.instance().info(f"dbt phase completed in {dbt_duration}s")

        parser = ParserFactory().build(project_file=discover.project_file, files=discover.files)
        try:
            project = parser.parse()
        except yaml.YAMLError as e:
            message = "\n"
            if hasattr(e, "problem_mark"):
                mark = e.problem_mark
                message = f"\n Error position: line:{mark.line+1} column:{mark.column+1}\n"
            raise click.ClickException(f"There was an error parsing the yml file(s):{message} {e}")

    if not project.defaults:
        project.defaults = Defaults()
    if default_source:
        project.defaults.source_name = default_source
    # Ensure output directory exists
    os.makedirs(output_dir, exist_ok=True)
    # Ensure dashboard directory exists
    dashboard_dir = get_dashboards_dir(output_dir)
    os.makedirs(dashboard_dir, exist_ok=True)

    return project



import os
from pathlib import Path
import re
import shutil
from flask import jsonify, request
from git import Repo
from visivo.commands.utils import create_source
from visivo.discovery.discover import Discover
from visivo.logger.logger import Logger
from visivo.models.dashboard import Dashboard
from visivo.models.defaults import Defaults
from visivo.models.include import Include
from visivo.models.item import Item
from visivo.models.models.sql_model import SqlModel
from visivo.models.project import Project
from visivo.models.row import Row
from visivo.models.source import CreateSourceRequest, SourceTypeEnum
from visivo.models.table import Table
from visivo.models.trace import Trace
from visivo.parsers.parser_factory import ParserFactory
from visivo.server.project_writer import ProjectWriter
from visivo.server.views.utils import create_example_dashboard, load_csv, write_project_file
from visivo.models.example_type import ExampleTypeEnum


def register_project_views(app, flask_app, output_dir):
    @app.route("/api/project/named_children/", methods=["GET"])
    def named_children():
        named_children = flask_app._project.named_child_nodes()
        if named_children:
            return jsonify(named_children)
        else:
            return jsonify({})

    @app.route("/api/project/project_file_path/", methods=["GET"])
    def project_file_path():
        project_file_path = flask_app._project.project_file_path
        if project_file_path:
            return jsonify(project_file_path)
        else:
            return jsonify({})

    @app.route("/api/project/write_changes/", methods=["POST"])
    def write_changes():
        data = request.get_json()
        if not data:
            return jsonify({"message": "No data provided"}), 400

        try:
            project_writer = ProjectWriter(data)
            project_writer.update_file_contents()
            project_writer.write()
            return jsonify({"message": "Changes written successfully"}), 200
        except Exception as e:
            Logger.instance().error(f"Error writing changes: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/project/load_example/", methods=["POST"])
    def load_example_project():
        """Load example project from GitHub"""
        from visivo.commands.init_phase import load_example_project as load_example

        data = request.get_json()
        project_name = data.get("project_name", "").strip()
        example_type = data.get("example_type", ExampleTypeEnum.github_releases.value)
        project_dir = data.get("project_dir", ".")

        if not project_name:
            return jsonify({"message": "Project name is required"}), 400

        # Pause the file watcher during cloning
        if flask_app.hot_reload_server:
            flask_app.hot_reload_server.pause_file_watcher()

        try:
            project_file_path = load_example(project_name, example_type, project_dir)

            # Initialize the project in Flask app
            if project_file_path:
                project_path = Path(project_dir)
                discover = Discover(working_dir=project_path, output_dir=None)

                try:
                    parser = ParserFactory().build(
                        project_file=discover.project_file, files=discover.files
                    )
                    project = parser.parse()
                    flask_app.project = project
                except Exception as e:
                    Logger.instance().error(f"Error parsing project: {str(e)}")

            return jsonify({"message": "Project created successfully"}), 200

        except Exception as e:
            Logger.instance().error(f"Error loading example project: {str(e)}")
            return jsonify({"message": f"Failed to load example project: {str(e)}"}), 500
        finally:
            if flask_app.hot_reload_server:
                Logger.instance().debug("Resuming file watcher after example load")
                flask_app.app.project = Project()
                flask_app.hot_reload_server.resume_file_watcher()
                flask_app.hot_reload_server.on_project_change()
                Logger.instance().debug("Resumed file watcher after example load")
                flask_app.hot_reload_server.socketio.emit("reload")

    @app.route("/api/project/init/", methods=["POST"])
    def init_project():
        data = request.get_json()
        project_name = data.get("project_name", "").strip()
        project_dir = data.get("project_dir", "").strip()

        if not project_name:
            return jsonify({"message": "Project name is required"}), 400

        project_path = (
            os.path.join(project_dir, "project.visivo.yml") if project_dir else "project.visivo.yml"
        )

        return jsonify({"message": "Project initialized", "project_file_path": project_path})

    @app.route("/api/source/create/", methods=["POST"])
    def create_source_api():
        form = request.form
        data = CreateSourceRequest(
            project_name=form.get("project_name", "Example Project"),
            source_name=form.get("source_name", "Example Source"),
            database=form.get("database", "Quickstart"),
            source_type=form.get("source_type"),
            host=form.get("host", ""),
            port=form.get("port"),
            username=form.get("username", ""),
            password=form.get("password", ""),
            account=form.get("account", ""),
            warehouse=form.get("warehouse", ""),
            credentials_base64=form.get("credentials_base64", ""),
            project=form.get("project", ""),
            dataset=form.get("dataset", ""),
            project_dir=form.get("project_dir", ""),
        )

        source = create_source(**data.model_dump())
        if isinstance(source, str):
            return jsonify({"message": source}), 400

        return jsonify({"message": "Source created", "source": source.model_dump()})

    @app.route("/api/source/upload/", methods=["POST"])
    def upload_file():
        file = request.files.get("file")
        source_type = request.form.get("source_type")
        source_name = request.form.get("source_name")
        project_dir = request.form.get("project_dir", "")

        if not file:
            return jsonify({"message": "File is required"}), 400

        file_path = os.path.join(project_dir, file.filename)
        file.save(file_path)

        table_name = re.sub(r"\W|^(?=\d)", "_", os.path.splitext(file.filename)[0])
        source = create_source(
            project_name="Temp",
            source_name=source_name,
            source_type=source_type,
            project_dir=project_dir,
        )

        with source.connect() as conn:
            conn.execute("INSTALL 'excel'")
            conn.execute("LOAD 'excel'")

            if source_type == SourceTypeEnum.csv:
                load_csv(conn, file_path, table_name)
            elif source_type == SourceTypeEnum.excel:
                conn.execute(
                    f"CREATE TABLE \"{table_name}\" AS SELECT * FROM read_xlsx('{file_path}')"
                )

            columns = conn.execute(f'PRAGMA table_info("{table_name}")').fetchall()
            quoted_cols = [f'"{col[1]}"' for col in columns]

            model = SqlModel(
                name="file_extract_model",
                sql=f'SELECT {", ".join(quoted_cols)} FROM "{table_name}"',
            )
            trace = Trace(
                name="file_extract_trace",
                model=model,
                columns={col[1]: f'"{col[1]}"' for col in columns},
                props={"type": "scatter"},
            )
            column_defs = [
                {"header": col[1].replace("_", " ").title(), "key": f"columns.{col[1]}"}
                for col in columns
            ]
            table = Table(
                name=f"{table_name}_table",
                traces=[trace],
                column_defs=[{"trace_name": trace.name, "columns": column_defs}],
            )
            dashboard = Dashboard(
                name="Uploaded File Dashboard", rows=[Row(items=[Item(table=table)])]
            )

            return jsonify({"message": "File uploaded", "dashboard": dashboard.model_dump()})

    @app.route("/api/project/finalize/", methods=["POST"])
    def finalize_project():
        data = request.get_json()
        project_name = data.get("project_name")
        project_dir = data.get("project_dir")
        sources = data.get("sources", [])
        dashboards = data.get("dashboards", [])
        include_example_dashboard = data.get("include_example_dashboard", True)

        includes = []

        if include_example_dashboard:
            includes.append(
                Include(
                    path="visivo-io/visivo.git@v0.0.1 -- test-projects/demo/dashboards/welcome.visivo.yml"
                )
            )

        if include_example_dashboard:
            dashboards.append(create_example_dashboard())

        project = Project(
            name=project_name,
            includes=includes,
            defaults=Defaults(source_name=sources[0]["name"] if sources else None),
            sources=sources,
            dashboards=dashboards,
        )

        write_project_file(project, project_dir)

        return jsonify({"message": "Project finalized"})



import urllib.parse
import webbrowser


def open_url(url: str) -> bool:
    """
    Tries to open the given URL in the default web browser.
    Returns True if webbrowser.open() claims success.
    """
    parsed = urllib.parse.urlparse(url)
    query_dict = dict(urllib.parse.parse_qsl(parsed.query))
    encoded_query = urllib.parse.urlencode(query_dict)
    safe_url = urllib.parse.urlunparse(
        (
            parsed.scheme,
            parsed.netloc,
            parsed.path,
            parsed.params,
            encoded_query,
            parsed.fragment,
        )
    )
    return webbrowser.open(safe_url)


def generate_success_html_response(
    base_url: str, timeout: int = 5, closePopUp: bool = False
) -> str:
    """
    Returns an HTML page that will redirect to base_url + '/profile' after `timeout` seconds.
    If closePopUp is True, it will attempt to close the tab instead.
    """
    redirect_url = f"{base_url}/profile"

    # Handle redirect or close the popup
    js_behavior = f"""
    <script>
      setTimeout(function() {{
        {"window.close();" if closePopUp else f"window.location.href = '{redirect_url}';"}
      }}, {timeout * 1000});
    </script>
    """

    return f"""
    <html>
      <head>
        <title>Authorization Successful</title>
        {js_behavior}
        <style>
        * {{
            margin: 0;
            padding: 0;
            box-sizing: border-box;
        }}
        body {{
            font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
            background: #F5F5F5;
            min-height: 100vh;
            display: flex;
            align-items: center;
            justify-content: center;
            position: relative;
            overflow: hidden;
        }}
        .container {{
            background: rgba(255, 255, 255, 0.95);
            backdrop-filter: blur(10px);
            border-radius: 20px;
            padding: 3rem 2.5rem;
            box-shadow: 0 10px 30px rgba(0, 0, 0, 0.1);
            max-width: 600px;
            width: 90%;
            text-align: center;
            position: relative;
            z-index: 2;
            border: 1px solid rgba(255, 255, 255, 0.2);
        }}
        .success-icon {{
            width: 80px;
            height: 80px;
            margin: 0 auto 1.5rem;
            background: linear-gradient(135deg, #10b981, #059669);
            border-radius: 50%;
            display: flex;
            align-items: center;
            justify-content: center;
        }}
        .checkmark {{
            width: 20px;
            height: 20px;
            color: white;
            font-weight: bold;
            font-size: 24px;
        }}
        h4 {{
            color: #713B57;
            font-size: 2.2rem;
            font-weight: 700;
            margin-bottom: 1rem;
            line-height: 1.3;
        }}
        .subtitle {{
            color: #713B57;
            font-size: 1.1rem;
            font-weight: 600;
            margin-bottom: 1.5rem;
            opacity: 0.8;
        }}
        .file-path {{
            background: #f8fafc;
            border: 2px solid #e2e8f0;
            border-radius: 8px;
            padding: 1rem;
            font-family: 'Courier New', monospace;
            font-size: 1rem;
            color: #374151;
            margin: 1.5rem 0;
            word-break: break-all;
        }}
        .description {{
            color: #64748b;
            font-size: 1rem;
            line-height: 1.6;
            margin-bottom: 2rem;
        }}
        .redirect-info {{
            background: #713B57;
            color: white;
            padding: 1rem 1.5rem;
            border-radius: 10px;
            font-size: 0.9rem;
            margin-top: 2rem;
        }}
        .loading-bar {{
            width: 100%;
            height: 4px;
            background: rgba(113, 59, 87, 0.2);
            border-radius: 2px;
            margin-top: 1rem;
            overflow: hidden;
        }}
        .loading-progress {{
            width: 0%;
            height: 100%;
            background: linear-gradient(90deg, #713B57, #875454);
            border-radius: 2px;
            animation: loadProgress {timeout}s linear forwards;
        }}
        @keyframes loadProgress {{
            to {{ width: 100%; }}
        }}
        @media (max-width: 768px) {{
            .container {{
                padding: 2rem 1.5rem;
                margin: 1rem;
            }}
            h1 {{
                font-size: 1.8rem;
            }}
            .file-path {{
                font-size: 0.9rem;
            }}
        }}
        </style>
      </head>
      <body>
        <div class="container">
            <div class="success-icon">
                <div class="checkmark">✓</div>
            </div>
            <h4>Authorization Successful!</h4>
            <p class="subtitle">Your development environment is now configured</p>
            <div class="file-path">~/.visivo/profile.yml</div>
            <p class="description">
                A secure token has been written to your development machine. This will enable seamless integration with your Visivo projects and development workflow.
            </p>
            <div class="redirect-info">
                <strong>{'Redirecting to local server please wait...' if closePopUp else 'Redirecting to your profile page...'}</strong>
                <br>
                You'll receive detailed instructions on creating new stages and projects.
                <div class="loading-bar">
                    <div class="loading-progress"></div>
                </div>
            </div>
        </div>
      </body>
    </html>
    """



from flask import jsonify
from visivo.logger.logger import Logger
from visivo.server.source_metadata import (
    check_source_connection,
    gather_source_metadata,
    get_database_schemas,
    get_schema_tables,
    get_source_databases,
    get_table_columns,
)


def register_source_views(app, flask_app, output_dir):
    @app.route("/api/project/sources_metadata/", methods=["GET"])
    def sources_metadata():
        try:
            metadata = gather_source_metadata(flask_app._project.sources)
            return jsonify(metadata)
        except Exception as e:
            Logger.instance().error(f"Error gathering source metadata: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/project/sources/<source_name>/test-connection/", methods=["GET"])
    def test_connection(source_name):
        """Test connection to a specific source."""
        try:
            result = check_source_connection(flask_app._project.sources, source_name)
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error testing connection for {source_name}: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/project/sources/<source_name>/databases/", methods=["GET"])
    def list_source_databases(source_name):
        """List databases for a specific source."""
        try:
            result = get_source_databases(flask_app._project.sources, source_name)
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing databases for {source_name}: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route(
        "/api/project/sources/<source_name>/databases/<database_name>/schemas/", methods=["GET"]
    )
    def list_database_schemas(source_name, database_name):
        """List schemas for a specific database."""
        try:
            result = get_database_schemas(flask_app._project.sources, source_name, database_name)
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing schemas: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route(
        "/api/project/sources/<source_name>/databases/<database_name>/tables/", methods=["GET"]
    )
    def list_database_tables(source_name, database_name):
        """List tables for a database (no schema)."""
        try:
            result = get_schema_tables(flask_app._project.sources, source_name, database_name)
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing tables: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route(
        "/api/project/sources/<source_name>/databases/<database_name>/schemas/<schema_name>/tables/",
        methods=["GET"],
    )
    def list_schema_tables(source_name, database_name, schema_name):
        """List tables for a specific schema."""
        try:
            result = get_schema_tables(
                flask_app._project.sources, source_name, database_name, schema_name
            )
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing tables: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route(
        "/api/project/sources/<source_name>/databases/<database_name>/tables/<table_name>/columns/",
        methods=["GET"],
    )
    def list_table_columns(source_name, database_name, table_name):
        """List columns for a table (no schema)."""
        try:
            result = get_table_columns(
                flask_app._project.sources, source_name, database_name, table_name
            )
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing columns: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route(
        "/api/project/sources/<source_name>/databases/<database_name>/schemas/<schema_name>/tables/<table_name>/columns/",
        methods=["GET"],
    )
    def list_schema_table_columns(source_name, database_name, schema_name, table_name):
        """List columns for a table in a specific schema."""
        try:
            result = get_table_columns(
                flask_app._project.sources, source_name, database_name, table_name, schema_name
            )
            if isinstance(result, tuple):  # Error response
                return jsonify(result[0]), result[1]
            return jsonify(result)
        except Exception as e:
            Logger.instance().error(f"Error listing columns: {str(e)}")
            return jsonify({"message": str(e)}), 500



from typing import Literal, Optional
from visivo.models.sources.source import BaseSource
from pydantic import Field
import duckdb
import click


class ExcelFileSource(BaseSource):
    type: Literal["xls"]
    file: str = Field(..., description="Path to the Excel file.")
    delimiter: Optional[str] = Field(",", description="Excel delimiter.")
    encoding: Optional[str] = Field("utf-8", description="Excel file encoding.")
    has_header: Optional[bool] = Field(True, description="Whether Excel has a header row.")

    def get_connection(self, read_only: bool = False):
        try:
            connection = duckdb.connect(":memory:")
            connection.execute(
                f"""
                CREATE VIEW "{self.name}" AS
                SELECT * FROM read_csv_auto('{self.file}', delim='{self.delimiter}', header={str(self.has_header).upper()})
                """
            )
            return connection
        except Exception as err:
            raise click.ClickException(
                f"Error connecting to Excel source '{self.name}'. Full Error: {str(err)}"
            )

    def read_sql(self, query: str):
        try:
            with self.connect(read_only=True) as connection:
                result = connection.execute(query)
                columns = [desc[0] for desc in result.description] if result.description else []
                rows = result.fetchall()
                return [dict(zip(columns, row)) for row in rows]
        except Exception as err:
            raise click.ClickException(
                f"Error executing query on Excel source '{self.name}': {str(err)}"
            )

    def connect(self, read_only: bool = False):
        return ExcelConnection(source=self, read_only=read_only)

    def get_dialect(self):
        return "duckdb"


class ExcelConnection:

    def __init__(self, source: ExcelFileSource, read_only: bool = False):
        self.source = source
        self.conn = None
        self.read_only = read_only

    def __enter__(self):
        self.conn = self.source.get_connection(read_only=self.read_only)
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()
            self.conn = None



"""
Utility functions for telemetry.
"""

import hashlib
from typing import Optional


def hash_project_name(project_name: Optional[str]) -> Optional[str]:
    """
    Hash a project name for privacy-preserving analytics.

    Uses SHA-256 with a salt to create a consistent but irreversible
    hash of the project name. This allows tracking unique projects
    without exposing actual project names.

    Args:
        project_name: The project name to hash

    Returns:
        str: Hexadecimal hash of the project name, or None if no name provided
    """
    if not project_name:
        return None

    # Use a fixed salt to ensure consistent hashing across runs
    # This salt makes it harder to reverse-engineer project names
    salt = "visivo-telemetry-v1"

    # Create hash
    hash_input = f"{salt}:{project_name}".encode("utf-8")
    hash_value = hashlib.sha256(hash_input).hexdigest()

    # Return first 16 characters for brevity (still plenty of entropy)
    return hash_value[:16]



"""
Flask middleware for automatic API telemetry tracking.
"""

import time
from flask import Flask, request, g
from .client import get_telemetry_client
from .events import APIEvent
from .config import is_telemetry_enabled


def init_telemetry_middleware(app: Flask, project=None):
    """
    Initialize telemetry middleware for a Flask app.

    Args:
        app: The Flask application
        project: Optional project object to check for telemetry settings
    """
    # Check if telemetry is enabled
    project_defaults = project.defaults if project else None
    telemetry_enabled = is_telemetry_enabled(project_defaults)

    if not telemetry_enabled:
        return

    # Get the telemetry client
    client = get_telemetry_client(enabled=True)

    # Hash the project name if available
    project_hash = None
    if project and hasattr(project, "name"):
        from .utils import hash_project_name

        project_hash = hash_project_name(project.name)

    @app.before_request
    def before_request():
        """Record the start time of the request."""
        g.start_time = time.time()

    @app.after_request
    def after_request(response):
        """Track the API request after it completes."""
        # Skip if no start time (shouldn't happen)
        if not hasattr(g, "start_time"):
            return response

        # Calculate duration
        duration_ms = int((time.time() - g.start_time) * 1000)

        # Create sanitized endpoint path (remove IDs and sensitive data)
        endpoint = request.endpoint or request.path
        if endpoint:
            # Replace common ID patterns with placeholders
            import re

            # UUID pattern
            endpoint = re.sub(
                r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}",
                "{id}",
                endpoint,
            )
            # Numeric IDs
            endpoint = re.sub(r"/\d+", "/{id}", endpoint)
            # Hash-like strings
            endpoint = re.sub(r"/[0-9a-fA-F]{32,}", "/{hash}", endpoint)

        # Track the event
        event = APIEvent.create(
            endpoint=endpoint,
            method=request.method,
            status_code=response.status_code,
            duration_ms=duration_ms,
            project_hash=project_hash,
        )

        client.track(event)

        return response

    @app.teardown_appcontext
    def teardown(error=None):
        """Ensure telemetry is flushed on app teardown."""
        # This is called when the app context tears down
        # We don't need to do anything here as our client handles this
        pass



"""
Telemetry client using PostHog SDK.
"""

import posthog
from typing import Optional
from .config import TELEMETRY_ENDPOINT, POSTHOG_API_KEY, POSTHOG_HOST, is_telemetry_enabled
from .events import BaseEvent


class TelemetryClient:
    """
    Telemetry client wrapper around PostHog SDK.

    This client maintains the same interface as the previous implementation
    but delegates to PostHog for actual event tracking.
    """

    def __init__(self, enabled: bool = True):
        """
        Initialize the telemetry client.

        Args:
            enabled: Whether telemetry is enabled
        """
        self.enabled = enabled
        self._initialized = False

        if self.enabled:
            self._initialize_posthog()

    def _initialize_posthog(self):
        """Initialize PostHog with our settings."""
        if self._initialized:
            return

        # Initialize PostHog with configuration
        posthog.api_key = POSTHOG_API_KEY
        posthog.host = POSTHOG_HOST

        # Configure PostHog settings for privacy
        posthog.disabled = not self.enabled
        posthog.personal_api_key = None  # Don't use personal API keys

        # Set batch settings similar to our previous implementation
        posthog.on_error = None  # Silently ignore errors
        posthog.debug = False
        posthog.send = True

        self._initialized = True

    def track(self, event: BaseEvent):
        """
        Track an event using PostHog.

        Args:
            event: The event to track
        """
        if not self.enabled:
            return

        try:
            event_dict = event.to_dict()

            # PostHog expects a specific format
            # Don't pass timestamp as string - PostHog will add its own
            posthog.capture(
                distinct_id=event_dict.get("machine_id", "anonymous"),
                event=event_dict.get("event_type", "unknown"),
                properties={
                    **event_dict.get("properties", {}),
                    "session_id": event_dict.get("session_id"),
                    "timestamp": event_dict.get("timestamp"),
                    # Override IP to prevent PII collection
                    "$ip": "0.0.0.0",
                },
            )
        except Exception:
            # Silently ignore all telemetry errors
            pass

    def flush(self, timeout: float = 2.0):
        """
        Flush any pending events.

        Args:
            timeout: Maximum time to wait for flush to complete
        """
        if not self.enabled:
            return

        try:
            posthog.flush()
        except Exception:
            # Silently ignore flush errors
            pass

    def shutdown(self):
        """Shutdown the telemetry client and send any remaining events."""
        if not self.enabled:
            return

        try:
            # PostHog handles shutdown automatically
            posthog.shutdown()
        except Exception:
            # Silently ignore shutdown errors
            pass


# Global telemetry client instance (created on first use)
_global_client: Optional[TelemetryClient] = None


def get_telemetry_client(enabled: bool = True) -> TelemetryClient:
    """
    Get the global telemetry client instance.

    Args:
        enabled: Whether telemetry should be enabled

    Returns:
        The global telemetry client
    """
    global _global_client
    if _global_client is None:
        _global_client = TelemetryClient(enabled=enabled)
    return _global_client



import platform

import urllib
from uuid import uuid4

from flask import jsonify, request

from visivo.logger.logger import Logger
from visivo.server.constants import VISIVO_HOST
from visivo.tokens.token_functions import get_existing_token, validate_and_store_token
from visivo.tokens.web_utils import generate_success_html_response
from visivo.server.store import background_jobs, background_jobs_lock

FLASK_DEFAULT_PORT = 8000


def register_auth_views(app, flask_app, output_dir):
    @app.route("/api/auth/status/", methods=["POST"])
    def authorize_status():
        existing_token = get_existing_token()
        if existing_token:
            return (
                jsonify(
                    {
                        "message": f"A token already exists in your profile: {existing_token}",
                        "token": existing_token,
                    }
                ),
                200,
            )

        return jsonify({"message": "UnAuthenticated user access"})

    @app.route("/api/auth/authorize-device-token/", methods=["POST"])
    def authorize_device_token():
        auth_id = str(uuid4())
        device_name = platform.node()
        redirect_url = f"http://localhost:{FLASK_DEFAULT_PORT}/api/auth/authorize-device-token/callback/{auth_id}/"

        params = {"redirect_url": redirect_url, "name": device_name}

        query_string = urllib.parse.urlencode(params)

        full_url = f"{VISIVO_HOST}/authorize-device?{query_string}"

        with background_jobs_lock:
            background_jobs[auth_id] = {
                "status": 202,
                "message": "Autheticating ...",
            }

        return jsonify(
            {
                "message": "Authentication initiated successfully",
                "auth_id": auth_id,
                "full_url": full_url,
            }
        )

    @app.route("/api/auth/authorize-device-token/callback/<auth_id>/", methods=["GET", "POST"])
    def authorize_device_callback(auth_id):
        token = request.args.get("token") if request.method == "GET" else None
        if not token:
            with background_jobs_lock:
                if auth_id in background_jobs:
                    background_jobs[auth_id] = {"message": "UnAuthorized access", "status": 401}
            return jsonify({"error": "Token not provided"}), 400

        Logger.instance().success("Received token via callback: " + token)
        validate_and_store_token(token)

        html_content = generate_success_html_response(VISIVO_HOST, timeout=5, closePopUp=True)

        with background_jobs_lock:
            if auth_id in background_jobs:
                background_jobs[auth_id] = {"message": "Authenticated", "status": 200}

        return html_content, 200



import hashlib
import os

from flask import Response, jsonify, request, send_from_directory

from visivo.logger.logger import Logger


def register_dashboard_views(app, flask_app, output_dir):
    @app.route("/api/dashboards/<dashboard_name>/", methods=["GET"])
    def get_dashboard_api(dashboard_name):
        """API endpoint for dashboard data"""
        try:
            dashboard_name_hash = hashlib.md5(dashboard_name.encode()).hexdigest()
            thumbnail_path = os.path.join("dashboards", f"{dashboard_name_hash}.png")
            thumbnail_exists = os.path.exists(os.path.join(output_dir, thumbnail_path))

            return jsonify(
                {
                    "id": dashboard_name,
                    "name": dashboard_name,
                    "signed_thumbnail_file_url": (
                        f"/api/dashboards/{dashboard_name_hash}.png/" if thumbnail_exists else None
                    ),
                }
            )
        except Exception as e:
            Logger.instance().error(f"Error fetching dashboard data: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/dashboards/<dashboard_name_hash>.png/", methods=["GET"])
    def get_thumbnail_api(dashboard_name_hash):
        try:
            thumbnail_path = os.path.join("dashboards", f"{dashboard_name_hash}.png")
            if not os.path.exists(os.path.join(output_dir, thumbnail_path)):
                Logger.instance().debug(f"Thumbnail not found at path: {thumbnail_path}")
                return Response(status=404)

            return send_from_directory(output_dir, thumbnail_path)
        except Exception as e:
            Logger.instance().error(f"Error retrieving thumbnail: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/dashboards/<dashboard_name_hash>.png/", methods=["POST"])
    def create_thumbnail_api(dashboard_name_hash):
        try:
            if "file" not in request.files:
                return jsonify({"message": "No file provided"}), 400

            file = request.files["file"]
            if file.filename == "" or not file.filename.endswith(".png"):
                return jsonify({"message": "Invalid file - must be a PNG"}), 400

            dashboard_dir = os.path.join(output_dir, "dashboards")
            os.makedirs(dashboard_dir, exist_ok=True)

            thumbnail_path = os.path.join(dashboard_dir, f"{dashboard_name_hash}.png")

            file.save(thumbnail_path)

            return jsonify(
                {
                    "signed_thumbnail_file_url": f"/api/dashboards/{dashboard_name_hash}.png/",
                }
            )

        except Exception as e:
            Logger.instance().error(f"Error creating thumbnail: {str(e)}")
            return jsonify({"message": str(e)}), 500



from typing import Literal, Optional, List, Any
from visivo.models.base.base_model import BaseModel
from visivo.models.sources.sqlalchemy_source import SqlalchemySource
from pydantic import Field, PrivateAttr
import click
from sqlalchemy import create_engine, event
from sqlalchemy.pool import NullPool
from visivo.logger.logger import Logger
from threading import Lock

attach_function_lock = Lock()

DuckdbType = Literal["duckdb"]


class DuckdbAttachment(BaseModel):
    schema_name: str = Field("Name of the schema to attach the source under.")
    source: "DuckdbSource" = Field(
        None,
        description="Local Duckdb database source to attach in the connection that will be available in the base SQL query.",
    )


class DuckdbSource(SqlalchemySource):
    # Cache both read-only and read-write engines
    _read_only_engine: Any = PrivateAttr(default=None)
    """
    DuckdbSources hold the connection information to DuckDB data sources.

    !!! example

        === "Simple"

            ``` yaml
                sources:
                  - name: duckdb_source
                    type: duckdb
                    database: local/file/database.db
                    connection_pool_size: 2
            ```

    !!! note

        Recommended environment variable use is covered in the [sources overview.](/topics/sources/)
    """

    type: DuckdbType
    connection_pool_size: Optional[int] = Field(
        1, description="The pool size that is used for this connection."
    )
    attach: Optional[List[DuckdbAttachment]] = Field(
        None,
        description="List of other local Duckdb database sources to attach in the connection that will be available in the base SQL query.",
    )

    def get_engine(self, read_only: bool = False):
        """Get engine with proper read_only support for DuckDB."""
        if read_only:
            if not self._read_only_engine:
                Logger.instance().debug(f"Creating read-only engine for Source: {self.name}")
                self._read_only_engine = create_engine(
                    self.url(), poolclass=NullPool, connect_args={"read_only": True}
                )

                @event.listens_for(self._read_only_engine, "connect")
                def connect_readonly(dbapi_connection, connection_record):
                    if self.after_connect:
                        cursor_obj = dbapi_connection.cursor()
                        cursor_obj.execute(self.after_connect)
                        cursor_obj.close()

            return self._read_only_engine
        else:
            # Create read-write engine with proper connection args
            if not self._engine:
                Logger.instance().debug(f"Creating read-write engine for Source: {self.name}")
                self._engine = create_engine(
                    self.url(), poolclass=NullPool, connect_args={"read_only": False}
                )

                @event.listens_for(self._engine, "connect")
                def connect_readwrite(dbapi_connection, connection_record):
                    if self.after_connect:
                        cursor_obj = dbapi_connection.cursor()
                        cursor_obj.execute(self.after_connect)
                        cursor_obj.close()

            return self._engine

    def safe_attach(self, connection, db_path, alias):
        with attach_function_lock:
            result = connection.execute(
                """
                SELECT count(*) as cnt
                FROM pragma_database_list
                WHERE name = ?
            """,
                [alias],
            ).fetchone()

            if result[0] == 0:
                connection.execute(f"ATTACH '{db_path}' AS {alias};")
                return True
            return False

    # Usage
    def get_connection(self, read_only: bool = False):
        """Return a DuckDBPyConnection using direct DuckDB connection with proper read_only support."""
        try:
            import duckdb
            import os

            Logger.instance().debug(f"Getting connection for {self.name}, read_only={read_only}")

            # Ensure database file exists for write operations
            if not read_only and not os.path.exists(self.database):
                Logger.instance().debug(
                    f"Database file {self.database} does not exist, creating it"
                )
                os.makedirs(os.path.dirname(self.database), exist_ok=True)
                # Create the database file
                temp_conn = duckdb.connect(self.database)
                temp_conn.close()

            # Connect directly to DuckDB with proper read_only flag
            Logger.instance().debug(f"Connecting to {self.database} with read_only={read_only}")
            connection = duckdb.connect(self.database, read_only=read_only)
            Logger.instance().debug(f"Direct DuckDB connection established for {self.name}")

            # Execute after_connect if specified
            if self.after_connect:
                connection.execute(self.after_connect)

            if self.attach:
                for attachment in self.attach:
                    if self.safe_attach(
                        connection, attachment.source.database, attachment.schema_name
                    ):
                        Logger.instance().debug(f"Database {attachment.schema_name} attached")
                    else:
                        Logger.instance().debug(
                            f"Database {attachment.schema_name} already attached"
                        )
            return connection
        except Exception as err:
            raise click.ClickException(
                f"Error connecting to source '{self.name}'. Ensure the database exists and the connection properties are correct. Full Error: {str(err)}"
            )

    def get_dialect(self):
        return "duckdb"

    def read_sql(self, query: str):
        try:
            with self.connect(read_only=True) as connection:
                # Execute query and get raw results
                result = connection.execute(query)

                # Get column names
                columns = [desc[0] for desc in result.description] if result.description else []

                # Fetch all rows
                rows = result.fetchall()

                # Convert to list of dictionaries
                data = []
                for row in rows:
                    row_dict = {}
                    for i, col in enumerate(columns):
                        row_dict[col] = row[i]
                    data.append(row_dict)

                return data
        except Exception as err:
            raise click.ClickException(f"Error executing query on source '{self.name}': {str(err)}")

    def connect(self, read_only: bool = False):
        """Create a context manager for DuckDB connections."""
        return DuckDBConnection(source=self, read_only=read_only)

    def connect_args(self):
        """Override to provide DuckDB-specific connection arguments for the default engine."""
        # The default cached engine (via super().get_engine()) is read-write
        # Read-only connections use a separate cached engine
        return {}

    def list_databases(self):
        """Return list of databases for DuckDB (includes attached databases)."""
        try:
            with self.connect(read_only=True) as connection:
                rows = connection.execute("PRAGMA database_list").fetchall()
                databases = []

                # DuckDB returns: (sequence_num, db_name, file_path)
                # For DuckDB, we always return 'main' as the primary database name
                # This provides consistency and avoids exposing file paths in the UI

                Logger.instance().debug(f"DuckDB PRAGMA database_list result: {rows}")

                # Always add 'main' for the primary database
                databases.append("main")

                # Add any attached databases (excluding temp/system)
                for row in rows:
                    db_name = row[1]
                    if db_name not in ["temp", "system", "main"] and row[2]:
                        # For attached databases, use their actual names
                        databases.append(db_name)

                Logger.instance().debug(f"DuckDB list_databases returning: {databases}")
                return databases
        except Exception as e:
            # Re-raise to allow proper error handling in UI
            raise e

    def dispose_engines(self):
        """Dispose of cached engines to release database locks."""
        if self._read_only_engine:
            self._read_only_engine.dispose()
            self._read_only_engine = None
        if self._engine:
            self._engine.dispose()
            self._engine = None

    def __del__(self):
        """Ensure engines are disposed when source is destroyed."""
        try:
            self.dispose_engines()
        except Exception:
            # Ignore errors during cleanup
            pass

    @classmethod
    def create_empty_database(cls, database_path: str):
        """Create an empty DuckDB database file."""
        import duckdb
        import os

        # Ensure directory exists
        os.makedirs(os.path.dirname(database_path), exist_ok=True)

        # Remove file if it exists (in case it's an invalid empty file)
        if os.path.exists(database_path):
            os.unlink(database_path)

        # Create empty database
        conn = duckdb.connect(database_path)
        conn.close()


class DuckDBConnection:
    def __init__(self, source: DuckdbSource, read_only: bool = False):
        self.source = source
        self.conn = None
        self.read_only = read_only

    def __enter__(self):
        self.conn = self.source.get_connection(read_only=self.read_only)
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()
            self.conn = None
        # Return None to propagate any exceptions
        return None



from flask import jsonify, request
from visivo.logger.logger import Logger


def register_worksheet_views(app, flask_app, output_dir):

    @app.route("/api/worksheet/", methods=["GET"])
    def list_worksheets():
        """List all worksheets."""
        try:
            worksheets = flask_app.worksheet_repo.list_worksheets()
            return jsonify(worksheets)
        except Exception as e:
            Logger.instance().error(f"Error listing worksheets: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/<worksheet_id>/", methods=["GET"])
    def get_worksheet(worksheet_id):
        """Get a specific worksheet."""
        try:
            worksheet = flask_app.worksheet_repo.get_worksheet(worksheet_id)
            if worksheet is None:
                return jsonify({"message": "Worksheet not found"}), 404
            return jsonify(worksheet)
        except Exception as e:
            Logger.instance().error(f"Error getting worksheet: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/", methods=["POST"])
    def create_worksheet():
        """Create a new worksheet."""
        try:
            data = request.get_json()
            if not data or "name" not in data:
                return jsonify({"message": "Name is required"}), 400

            result = flask_app.worksheet_repo.create_worksheet(
                name=data["name"],
                query=data.get("query", ""),
                selected_source=data.get("selected_source"),
            )
            return jsonify(result), 201
        except Exception as e:
            Logger.instance().error(f"Error creating worksheet: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/<worksheet_id>/", methods=["PUT"])
    def update_worksheet(worksheet_id):
        """Update a worksheet."""
        try:
            data = request.get_json()
            if not data:
                return jsonify({"message": "No update data provided"}), 400

            success = flask_app.worksheet_repo.update_worksheet(worksheet_id, data)
            if not success:
                return jsonify({"message": "Worksheet not found"}), 404

            return jsonify({"message": "Worksheet updated successfully"})
        except Exception as e:
            Logger.instance().error(f"Error updating worksheet: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/<worksheet_id>/", methods=["DELETE"])
    def delete_worksheet(worksheet_id):
        """Delete a worksheet."""
        try:
            success = flask_app.worksheet_repo.delete_worksheet(worksheet_id)
            if not success:
                return jsonify({"message": "Worksheet not found"}), 404
            return jsonify({"message": "Worksheet deleted successfully"})
        except Exception as e:
            Logger.instance().error(f"Error deleting worksheet: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/session/", methods=["GET"])
    def get_session_state():
        """Get the current session state."""
        try:
            worksheets = flask_app.worksheet_repo.list_worksheets()
            return jsonify([w["session_state"] for w in worksheets])
        except Exception as e:
            Logger.instance().error(f"Error getting session state: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/worksheet/session/", methods=["PUT"])
    def update_session_state():
        """Update the session state."""
        try:
            data = request.get_json()
            if not isinstance(data, list):
                return jsonify({"message": "Expected array of session states"}), 400

            success = flask_app.worksheet_repo.update_session_states(data)
            if not success:
                return jsonify({"message": "Failed to update session state"}), 500

            return jsonify({"message": "Session state updated successfully"})
        except Exception as e:
            Logger.instance().error(f"Error updating session state: {str(e)}")
            return jsonify({"message": str(e)}), 500



import os
from pathlib import Path
from time import time

import click

from visivo.commands.options import (
    dag_filter,
    output_dir,
    working_dir,
    source,
    port,
    threads,
    dbt_profile,
    dbt_target,
    skip_compile,
    new,
)
from visivo.discovery.discover import Discover
from visivo.models.defaults import Defaults
from visivo.models.project import Project
from visivo.commands.serve_phase import serve_phase
from visivo.commands.parse_project_phase import parse_project_phase
from visivo.logger.logger import Logger


@click.command()
@source
@working_dir
@output_dir
@dag_filter
@port
@threads
@skip_compile
@dbt_profile
@dbt_target
@new
@click.pass_context
def serve(
    ctx,
    output_dir,
    working_dir,
    source,
    port,
    dag_filter,
    threads,
    skip_compile,
    dbt_profile,
    dbt_target,
    new,
    project_dir,
    pd,
):
    start_time = time()
    logger = Logger.instance()
    server_url = f"http://localhost:{port}"
    is_default_working_dir = ctx.obj.get("is_default_working_dir")

    if is_default_working_dir and not new:
        discover = Discover(working_dir=working_dir, output_dir=output_dir)
        new = not discover.project_file_exists

    # Handle new project creation
    if new:
        skip_compile = True
        project = Project(
            name="Quickstart Visivo", sources=[], models=[], traces=[], charts=[], dashboards=[]
        )

        project.defaults = project.defaults or Defaults()
        if source:
            project.defaults.source_name = source

        final_project_dir = pd or project_dir or os.path.abspath(".")
        if os.path.exists(final_project_dir) and final_project_dir != os.getcwd():
            logger.error(f"Project already exists at '{final_project_dir}'")
            raise click.ClickException("Project directory already exists.")

        project.project_dir = final_project_dir
        working_dir = final_project_dir
        os.makedirs(output_dir, exist_ok=True)
        os.makedirs(working_dir, exist_ok=True)

    else:
        project = parse_project_phase(
            working_dir=working_dir,
            output_dir=output_dir,
            default_source=source,
            dbt_profile=dbt_profile,
            dbt_target=dbt_target,
        )

    # Start the development server
    server, on_change, on_ready = serve_phase(
        output_dir=output_dir,
        working_dir=working_dir,
        default_source=source,
        dag_filter=dag_filter,
        threads=threads,
        skip_compile=skip_compile,
        project=project,
        server_url=server_url,
        new=new,
    )

    logger.info(f"Initial build completed in {round(time() - start_time, 2)}s")
    logger.info(f"Server running at {server_url}")

    server.serve(
        host="0.0.0.0",
        port=port,
        on_change_callback=on_change,
        on_server_ready=on_ready,
    )



import click
from visivo.commands.options import project_dir
from visivo.models.example_type import ExampleTypeEnum


@click.command()
@project_dir
@click.option(
    "--example",
    type=click.Choice([e.value for e in ExampleTypeEnum]),
    help="Load an example project from GitHub",
    default=None,
)
def init(project_dir, example):
    """
    Initialize a new Visivo project.

    By default, creates a simple project.visivo.yml file with a project name.
    Use --example to load an example project from GitHub.
    """
    from visivo.logger.logger import Logger

    Logger.instance().debug("Init")

    from visivo.commands.init_phase import init_phase

    init_phase(project_dir, example)

    Logger.instance().success("Done")



from typing import Union
import click
import os
from pathlib import Path
from visivo.logger.logger import Logger
from visivo.models.source import SourceTypeEnum
from visivo.models.sources.bigquery_source import BigQuerySource, BigQueryType
from visivo.models.sources.duckdb_source import DuckdbSource, DuckdbType
from visivo.models.sources.mysql_source import MysqlSource, MysqlType
from visivo.models.sources.postgresql_source import PostgresqlSource, PostgresqlType
from visivo.models.sources.snowflake_source import SnowflakeSource, SnowflakeType
from visivo.models.sources.sqlite_source import SqliteSource, SqliteType
from visivo.parsers.file_names import PROFILE_FILE_NAME
from visivo.utils import load_yaml_file


def get_source_types():
    from typing import get_args
    from visivo.models.sources.sqlite_source import SqliteType
    from visivo.models.sources.postgresql_source import PostgresqlType
    from visivo.models.sources.mysql_source import MysqlType
    from visivo.models.sources.snowflake_source import SnowflakeType
    from visivo.models.sources.bigquery_source import BigQueryType
    from visivo.models.sources.duckdb_source import DuckdbType

    sqlite_type = get_args(SqliteType)[0]
    postgresql_type = get_args(PostgresqlType)[0]
    mysql_type = get_args(MysqlType)[0]
    snowflake_type = get_args(SnowflakeType)[0]
    bigquery_type = get_args(BigQueryType)[0]
    duckdb_type = get_args(DuckdbType)[0]
    return [postgresql_type, mysql_type, sqlite_type, snowflake_type, bigquery_type, duckdb_type]


def get_profile_token(profile_file):
    profile_token = os.getenv("VISIVO_TOKEN")
    if profile_token:
        return profile_token

    profile = None
    if profile_file:
        profile = load_yaml_file(profile_file)

    if not profile or "token" not in profile:
        raise click.ClickException(
            f"{PROFILE_FILE_NAME} not present or token not present in {PROFILE_FILE_NAME}"
        )
    return profile["token"]


def get_profile_file(home_dir=os.path.expanduser("~")):
    return Path(f"{home_dir}/.visivo/{PROFILE_FILE_NAME}")


def create_file_database(url, output_dir: str):
    from sqlalchemy import create_engine, MetaData, Table, Integer, Column, insert

    if output_dir != "":
        os.makedirs(output_dir, exist_ok=True)
    engine = create_engine(url, echo=True)
    metadata_obj = MetaData()
    table = Table(
        "test_table",
        metadata_obj,
        Column("x", Integer),
        Column("y", Integer),
    )
    second_table = Table(
        "second_test_table",
        metadata_obj,
        Column("x", Integer),
        Column("y", Integer),
    )
    metadata_obj.create_all(engine)
    for v in [[1, 1], [2, 1], [3, 2], [4, 3], [5, 5], [6, 8]]:
        with engine.connect() as connection:
            connection.execute(insert(table).values(x=v[0], y=v[1]))
            connection.execute(insert(second_table).values(x=v[0], y=v[1] * 2))
            connection.commit()


def create_project_path(project_dir=None) -> Union[str, None]:
    if not project_dir:
        return "Project name must be provided"

    if Path(project_dir).exists():
        return f"'{project_dir}' directory already exists"

    os.makedirs(project_dir.strip(), exist_ok=True)
    return None


def create_source(
    project_name: str = "",
    source_name: str = "Example Source",
    project_dir: str = "",
    database: str = "",
    source_type: Union[
        SqliteType, PostgresqlType, MysqlType, SnowflakeType, BigQueryType, DuckdbType
    ] = None,
    host: str = "",
    port: int = None,
    username: str = "",
    password: str = "",
    account: str = "",
    warehouse: str = "",
    credentials_base64: str = "",
    project: str = "",
    dataset: str = "",
) -> Union[
    SqliteSource,
    PostgresqlSource,
    MysqlSource,
    SnowflakeSource,
    BigQuerySource,
    DuckdbSource,
    str,
    None,
]:
    project_dir = project_dir.strip()
    default_db_name = f"{database or 'local'}.db"

    local_db_path = Path(project_dir) / default_db_name if project_dir else default_db_name
    env_path = Path(project_dir) / ".env" if project_dir else Path(".env")

    def write_env(var: str, value: str):
        env_path.write_text(f"{var}={value}")

    if source_type in {
        SourceTypeEnum.duckdb,
        SourceTypeEnum.csv,
        SourceTypeEnum.excel,
    }:
        # All use duckdb engine under the hood
        source = DuckdbSource(
            name=source_name, database=str(local_db_path), type=SourceTypeEnum.duckdb
        )
        if source_type == SourceTypeEnum.duckdb:
            create_file_database(source.url(), project_dir)

        write_env("DB_PASSWORD", "EXAMPLE_password_l0cation")
        return source

    if source_type == SourceTypeEnum.sqlite:
        source = SqliteSource(name=source_name, database=str(local_db_path), type=source_type)
        create_file_database(source.url(), project_dir)
        write_env("DB_PASSWORD", "EXAMPLE_password_l0cation")
        return source

    if source_type == SourceTypeEnum.postgresql:
        write_env("DB_PASSWORD", "postgres")
        source = PostgresqlSource(
            name=source_name,
            host=host,
            port=port or 5432,
            database=database,
            type=source_type,
            password="postgres",
            username=username,
        )
        return source

    if source_type == SourceTypeEnum.mysql:
        write_env("DB_PASSWORD", password)
        return MysqlSource(
            name=source_name,
            host=host,
            database=database,
            type=source_type,
            password=password,
            username=username,
        )

    if source_type == SourceTypeEnum.snowflake:
        write_env("DB_PASSWORD", password)
        return SnowflakeSource(
            name=source_name,
            database=database,
            type=source_type,
            password=password,
            username=username,
            account=account,
            warehouse=warehouse,
        )

    if source_type == SourceTypeEnum.bigquery:
        write_env("DB_PASSWORD", credentials_base64)
        return BigQuerySource(
            name=source_name,
            project=project,
            database=dataset,
            type=source_type,
            credentials_base64=credentials_base64,
        )

    return None



"""
Telemetry configuration and opt-out logic.
"""

import os
from pathlib import Path
from typing import Optional
import yaml

# Default telemetry endpoint - this should be configured based on your backend
TELEMETRY_ENDPOINT = os.getenv("VISIVO_TELEMETRY_ENDPOINT", "https://telemetry.visivo.io/v1/events")
TELEMETRY_TIMEOUT = 1.0  # Maximum time to wait for telemetry requests

# PostHog configuration
# This API key is for the Visivo project on PostHog Cloud
# Users can override this by setting VISIVO_POSTHOG_API_KEY environment variable
# or use their own self-hosted PostHog instance
POSTHOG_API_KEY = os.getenv(
    "VISIVO_POSTHOG_API_KEY", "phc_DaLOz39kD2u4ZFNi6aXQuA7ncmnbAGoE8dLZc2z7Agj"
)
POSTHOG_HOST = os.getenv("VISIVO_POSTHOG_HOST", "https://app.posthog.com")


def _check_env_disabled() -> bool:
    """Check if telemetry is disabled via environment variable."""
    return os.getenv("VISIVO_TELEMETRY_DISABLED", "").lower() in ("true", "1", "yes")


def _check_global_config_disabled() -> bool:
    """Check if telemetry is disabled in global config file."""
    config_path = Path.home() / ".visivo" / "config.yml"
    if not config_path.exists():
        return False

    try:
        with open(config_path, "r") as f:
            config = yaml.safe_load(f) or {}
            return config.get("telemetry_enabled", True) is False
    except Exception:
        # If we can't read the config, assume telemetry is enabled
        return False


def is_telemetry_enabled(project_defaults: Optional[object] = None) -> bool:
    """
    Check if telemetry is enabled based on all configuration sources.

    Args:
        project_defaults: Optional Defaults object from the project

    Returns:
        bool: True if telemetry is enabled, False otherwise
    """
    # Check environment variable first (highest priority)
    if _check_env_disabled():
        return False

    # Check project defaults if provided
    if project_defaults and hasattr(project_defaults, "telemetry_enabled"):
        if project_defaults.telemetry_enabled is False:
            return False

    # Check global config file
    if _check_global_config_disabled():
        return False

    # Telemetry is enabled by default
    return True


def is_ci_environment() -> bool:
    """
    Detect if we're running in a CI/CD environment.

    Checks for common CI environment variables that indicate
    we're running in an automated environment rather than
    on a developer's machine.

    Returns:
        bool: True if running in CI/CD, False otherwise
    """
    # Common CI environment variables
    ci_env_vars = [
        "CI",  # Generic CI indicator (GitHub Actions, GitLab CI, CircleCI, etc.)
        "CONTINUOUS_INTEGRATION",  # Generic
        "GITHUB_ACTIONS",  # GitHub Actions
        "GITLAB_CI",  # GitLab CI
        "CIRCLECI",  # CircleCI
        "JENKINS_HOME",  # Jenkins
        "JENKINS_URL",  # Jenkins
        "TEAMCITY_VERSION",  # TeamCity
        "TRAVIS",  # Travis CI
        "BUILDKITE",  # Buildkite
        "DRONE",  # Drone
        "BITBUCKET_BUILD_NUMBER",  # Bitbucket Pipelines
        "SEMAPHORE",  # Semaphore CI
        "APPVEYOR",  # AppVeyor
        "WERCKER",  # Wercker
        "MAGNUM",  # Magnum CI
        "MINT",  # Mint (rwx)
        "CODEBUILD_BUILD_ID",  # AWS CodeBuild
        "TF_BUILD",  # Azure DevOps
        "SYSTEM_TEAMFOUNDATIONCOLLECTIONURI",  # Azure DevOps
    ]

    # Check if any CI environment variable is set
    for var in ci_env_vars:
        if os.getenv(var):
            return True

    # Additional heuristics for container environments
    # Check if running in Docker
    if os.path.exists("/.dockerenv"):
        return True

    # Check for Kubernetes
    if os.getenv("KUBERNETES_SERVICE_HOST"):
        return True

    return False



"""
Telemetry module for Visivo CLI and API usage tracking.

This module provides anonymous usage telemetry to help improve Visivo.
Telemetry can be disabled via:
- Environment variable: VISIVO_TELEMETRY_DISABLED=true
- Project config: telemetry_enabled: false in defaults
- Global config: ~/.visivo/config.yml with telemetry_enabled: false
"""

from .client import TelemetryClient, get_telemetry_client
from .config import is_telemetry_enabled
from .context import get_telemetry_context

__all__ = [
    "TelemetryClient",
    "get_telemetry_client",
    "is_telemetry_enabled",
    "get_telemetry_context",
]



"""
Telemetry context for sharing metrics across different parts of the application.
"""

from typing import Dict, Any, Optional
import threading


class TelemetryContext:
    """
    Thread-safe context for storing telemetry data that needs to be
    collected from different parts of the application.
    """

    def __init__(self):
        self._data: Dict[str, Any] = {}
        self._lock = threading.Lock()

    def set(self, key: str, value: Any):
        """Set a value in the context."""
        with self._lock:
            self._data[key] = value

    def get(self, key: str, default: Any = None) -> Any:
        """Get a value from the context."""
        with self._lock:
            return self._data.get(key, default)

    def update(self, data: Dict[str, Any]):
        """Update multiple values at once."""
        with self._lock:
            self._data.update(data)

    def get_all(self) -> Dict[str, Any]:
        """Get all context data."""
        with self._lock:
            return self._data.copy()

    def clear(self):
        """Clear all context data."""
        with self._lock:
            self._data.clear()


# Global telemetry context instance
_context = TelemetryContext()


def get_telemetry_context() -> TelemetryContext:
    """Get the global telemetry context."""
    return _context



"""
Lightweight metric collection utilities for Visivo telemetry.
"""

from typing import Dict, Optional
from visivo.models.project import Project


def collect_project_metrics(project: Project) -> Dict[str, int]:
    """
    Collect high-level metrics from a project.

    This function is designed to be very fast and only collects
    counts of top-level objects without any expensive operations.

    Args:
        project: The Visivo project to collect metrics from

    Returns:
        Dictionary of object type to count
    """
    return {
        "sources": len(project.sources) if project.sources else 0,
        "models": len(project.models) if project.models else 0,
        "traces": len(project.traces) if project.traces else 0,
        "tables": len(project.tables) if project.tables else 0,
        "charts": len(project.charts) if project.charts else 0,
        "dashboards": len(project.dashboards) if project.dashboards else 0,
        "alerts": len(project.alerts) if project.alerts else 0,
        "selectors": len(project.selectors) if project.selectors else 0,
        "destinations": len(project.destinations) if project.destinations else 0,
    }


def count_filtered_jobs(project_dag, dag_filter: str) -> int:
    """
    Count the number of jobs that will be run based on the DAG filter.

    This is a lightweight operation that just counts the filtered DAGs
    without executing them.

    Args:
        project_dag: The project DAG object
        dag_filter: The filter string for selecting jobs

    Returns:
        Number of jobs that match the filter
    """
    try:
        # The filter_dag method returns an iterator of job DAGs
        # We just count them without materializing the full list
        return sum(1 for _ in project_dag.filter_dag(dag_filter))
    except Exception:
        # If anything goes wrong, return -1 to indicate error
        return -1



import os
import subprocess
from flask import jsonify, request
from visivo.server.text_editors import get_editor_configs


def register_editor_views(app, flask_app, output_dir):
    @app.route("/api/editors/installed/", methods=["GET"])
    def get_installed_editors():
        editors, platform = get_editor_configs()

        # Check which editors are installed
        installed_editors = []
        for editor in editors:
            # Skip editors that don't support this platform
            if not editor["paths"][platform]:
                continue

            for path in editor["paths"][platform]:
                if os.path.exists(path):
                    # Only send back safe information to the client
                    installed_editors.append({"name": editor["name"], "id": editor["id"]})
                    break

        return jsonify(installed_editors)

    @app.route("/api/editors/open/", methods=["POST"])
    def open_in_editor():
        data = request.get_json()
        if not data or "editorId" not in data or "filePath" not in data:
            return jsonify({"error": "Missing required parameters"}), 400

        editor_id = data["editorId"]
        file_path = data["filePath"]

        # Validate file path exists
        if not os.path.exists(file_path):
            return jsonify({"error": "File not found"}), 404

        # Get editor configurations
        editors, platform = get_editor_configs()

        # Find the selected editor
        editor_config = next((e for e in editors if e["id"] == editor_id), None)
        if not editor_config or not editor_config["commands"][platform]:
            return jsonify({"error": "Invalid editor for this platform"}), 400

        try:
            # Get the command for the selected editor
            command = editor_config["commands"][platform]

            # Special handling for VS Code on macOS
            if platform == "mac" and editor_id == "vscode":
                if command[0] == "open":
                    # When using 'open' command, file path goes at the end
                    subprocess.Popen(command + ["--", file_path])
                    subprocess.call(["wmctrl", "-a", file_path])
                else:
                    # When using 'code' command directly
                    subprocess.Popen(command + [file_path])
            else:
                # Normal handling for other editors
                subprocess.Popen(command + [file_path])

            return jsonify({"message": "File opened successfully"}), 200
        except Exception as e:
            return jsonify({"error": str(e)}), 500



from visivo.models.sources.sqlalchemy_source import SqlalchemySource
from visivo.logger.logger import Logger
from sqlalchemy import text
from typing import Optional, Tuple, Any, List


# Helper functions to reduce duplication


def _find_source(sources: List[Any], source_name: str) -> Optional[SqlalchemySource]:
    """Find a SqlalchemySource by name."""
    for src in sources:
        if isinstance(src, SqlalchemySource) and src.name == source_name:
            return src
    return None


def _get_engine_with_read_only(src: SqlalchemySource):
    """Get engine with read_only parameter if supported."""
    if hasattr(src, "get_engine") and "read_only" in src.get_engine.__code__.co_varnames:
        return src.get_engine(read_only=True)
    else:
        return src.get_engine()


def _switch_database_context(conn, engine, database_name: str, src_database: str):
    """Switch database context if needed for MySQL/Snowflake."""
    from sqlalchemy import inspect

    dialect = engine.dialect.name
    if dialect.startswith(("mysql", "snowflake")) and database_name != src_database:
        conn.execute(text(f"USE {database_name}"))
        return inspect(engine)  # Return new inspector after context switch
    return None


def _build_error_response(source_name: str, error: Exception, **kwargs) -> dict:
    """Build standardized error response."""
    response = {
        "source": source_name,
        "error": str(error),
    }
    response.update(kwargs)
    return response


def _source_not_found_error(source_name: str) -> Tuple[dict, int]:
    """Standard response for source not found."""
    return {"error": f"Source '{source_name}' not found"}, 404


# Main functions refactored


def check_source_connection(sources, source_name):
    """Test connection to a specific source."""
    src = _find_source(sources, source_name)
    if not src:
        return _source_not_found_error(source_name)

    try:
        Logger.instance().info(f"Testing connection for source: {source_name}")
        engine = _get_engine_with_read_only(src)

        # Test connection with a simple query - all dialects support SELECT 1
        with engine.connect() as conn:
            conn.execute(text("SELECT 1"))

        return {"source": source_name, "status": "connected"}

    except Exception as e:
        Logger.instance().debug(f"Connection test failed for {source_name}: {e}")
        return {"source": source_name, "status": "connection_failed", "error": str(e)}


def get_source_databases(sources, source_name):
    """Return list of databases for a specific source."""
    src = _find_source(sources, source_name)
    if not src:
        return _source_not_found_error(source_name)

    try:
        Logger.instance().info(f"Fetching databases for source: {source_name}")
        databases = src.list_databases()
        return {
            "source": source_name,
            "databases": [{"name": db} for db in databases],
            "status": "connected",
        }
    except Exception as e:
        Logger.instance().debug(f"Error fetching databases for {source_name}: {e}")
        return {
            "source": source_name,
            "databases": [],
            "status": "connection_failed",
            "error": str(e),
        }


def get_database_schemas(sources, source_name, database_name):
    """Return list of schemas for a specific database."""
    src = _find_source(sources, source_name)
    if not src:
        return _source_not_found_error(source_name)

    try:
        Logger.instance().info(f"Fetching schemas for {source_name}.{database_name}")
        from sqlalchemy import inspect

        engine = _get_engine_with_read_only(src)
        inspector = inspect(engine)

        # Switch database context if needed
        with engine.connect() as conn:
            new_inspector = _switch_database_context(conn, engine, database_name, src.database)
            if new_inspector:
                inspector = new_inspector

            # Get schemas
            try:
                schemas = inspector.get_schema_names()
                return {
                    "source": source_name,
                    "database": database_name,
                    "schemas": [{"name": s} for s in schemas] if schemas else None,
                    "has_schemas": bool(schemas),
                }
            except Exception:
                # Database doesn't support schemas
                return {
                    "source": source_name,
                    "database": database_name,
                    "schemas": None,
                    "has_schemas": False,
                }

    except Exception as e:
        Logger.instance().debug(f"Error fetching schemas for {source_name}.{database_name}: {e}")
        return _build_error_response(source_name, e, database=database_name), 500


def get_schema_tables(sources, source_name, database_name, schema_name=None):
    """Return list of tables for a specific database/schema."""
    src = _find_source(sources, source_name)
    if not src:
        return _source_not_found_error(source_name)

    try:
        Logger.instance().info(
            f"Fetching tables for {source_name}.{database_name}.{schema_name or 'default'}"
        )
        from sqlalchemy import inspect

        engine = _get_engine_with_read_only(src)
        inspector = inspect(engine)

        # Switch database context if needed
        with engine.connect() as conn:
            new_inspector = _switch_database_context(conn, engine, database_name, src.database)
            if new_inspector:
                inspector = new_inspector

            # Get tables
            tables = inspector.get_table_names(schema=schema_name)
            return {
                "source": source_name,
                "database": database_name,
                "schema": schema_name,
                "tables": [{"name": t} for t in tables],
            }

    except Exception as e:
        Logger.instance().debug(f"Error fetching tables: {e}")
        return (
            _build_error_response(source_name, e, database=database_name, schema=schema_name),
            500,
        )


def get_table_columns(sources, source_name, database_name, table_name, schema_name=None):
    """Return list of columns for a specific table."""
    src = _find_source(sources, source_name)
    if not src:
        return _source_not_found_error(source_name)

    try:
        Logger.instance().info(
            f"Fetching columns for {source_name}.{database_name}.{schema_name or 'default'}.{table_name}"
        )
        from sqlalchemy import inspect

        engine = _get_engine_with_read_only(src)
        inspector = inspect(engine)

        # Switch database context if needed
        with engine.connect() as conn:
            new_inspector = _switch_database_context(conn, engine, database_name, src.database)
            if new_inspector:
                inspector = new_inspector

            # Get columns
            columns = inspector.get_columns(table_name, schema=schema_name)
            return {
                "source": source_name,
                "database": database_name,
                "schema": schema_name,
                "table": table_name,
                "columns": [{"name": c["name"], "type": str(c["type"])} for c in columns],
            }

    except Exception as e:
        Logger.instance().debug(f"Error fetching columns: {e}")
        return (
            _build_error_response(
                source_name, e, database=database_name, schema=schema_name, table=table_name
            ),
            500,
        )


def gather_source_metadata(sources):
    """Return metadata for all provided sources."""
    data = {"sources": []}
    for src in sources:
        if isinstance(src, SqlalchemySource):
            try:
                Logger.instance().info(f"Attempting introspection for source: {src.name}")
                metadata = src.introspect()

                # Check if introspection returned an error result
                if "error" in metadata:
                    Logger.instance().debug(
                        f"Error during introspection for {src.name}: {metadata['error']}"
                    )
                    failed_metadata = {
                        "name": src.name,
                        "type": src.type,
                        "status": "connection_failed",
                        "error": metadata["error"],
                        "databases": [],
                    }
                    data["sources"].append(failed_metadata)
                else:
                    metadata["status"] = "connected"
                    data["sources"].append(metadata)
                    Logger.instance().info(f"Successfully introspected source: {src.name}")

            except Exception as e:
                Logger.instance().debug(f"Error during introspection for {src.name}: {e}")
                # Include failed source with error status
                failed_metadata = {
                    "name": src.name,
                    "type": src.type,
                    "status": "connection_failed",
                    "error": str(e),
                    "databases": [],
                }
                data["sources"].append(failed_metadata)
    return data



import json
import os
from threading import Thread
from uuid import uuid4
from flask import copy_current_request_context, jsonify, request
import requests
from visivo.logger.logger import Logger
from visivo.server.constants import VISIVO_HOST
from visivo.tokens.token_functions import get_existing_token
from visivo.server.store import background_jobs, background_jobs_lock


def register_cloud_views(app, flask_app, output_dir):
    @app.route("/api/cloud/stages/", methods=["GET"])
    def cloud_stages():
        token = get_existing_token()

        json_headers = {
            "content-type": "application/json",
            "Authorization": f"Api-Key {token}",
        }

        response = requests.get(f"{VISIVO_HOST}/api/stages/", headers=json_headers)

        if response.status_code == 200:
            return jsonify({"message": "Stages fetched successfully", "stages": response.json()})

        if response.status_code == 401:
            return jsonify({"message": "UnAuthorized access", "stage": response.json()}), 401

        return jsonify({"message": "Something went wrong!", "stage": response.json()}), 500

    @app.route("/api/cloud/stages/", methods=["POST"])
    def create_cloud_stages():
        data = request.get_json()
        name = data.get("name", "")

        token = get_existing_token()

        if name == "":
            return jsonify({"message": "Name is required"}), 400

        json_headers = {
            "content-type": "application/json",
            "Authorization": f"Api-Key {token}",
        }

        body = {
            "name": name,
        }

        response = requests.post(
            f"{VISIVO_HOST}/api/stages/", data=json.dumps(body), headers=json_headers
        )

        if response.status_code == 201:
            return jsonify({"message": "Stages fetched successfully", "stage": response.json()})

        if response.status_code == 401:
            return jsonify({"message": "UnAuthorized access", "stage": response.json()}), 401

        return jsonify({"message": "Something went wrong!", "stage": response.json()}), 500

    def deploy(stage: str, deploy_id):
        from visivo.commands.deploy_phase import deploy_phase

        deploy_phase(
            user_dir=os.path.expanduser("~"),
            working_dir=flask_app._working_dir,
            output_dir=output_dir,
            stage=stage,
            host=VISIVO_HOST,
            deploy_id=deploy_id,
        )

    @app.route("/api/cloud/deploy/", methods=["POST"])
    def cloud_deploy():
        data = request.get_json()
        stage = data.get("name")
        deploy_id = str(uuid4())

        with background_jobs_lock:
            background_jobs[deploy_id] = {
                "status": 200,
                "messages": "",
                "project_url": None,
            }

        @copy_current_request_context
        def deploy_with_context(stage, deploy_id):
            deploy(stage, deploy_id)
            with background_jobs_lock:
                background_jobs[deploy_id]["status"] = 201

        thread = Thread(target=deploy_with_context, args=(stage, deploy_id), daemon=True)
        thread.start()

        return jsonify({"message": "Deployment initiated successfully", "deploy_id": deploy_id})

    @app.route("/api/cloud/job/status/<deploy_id>/", methods=["GET"])
    def get_job_status(deploy_id):
        with background_jobs_lock:
            job = background_jobs.get(deploy_id)
            if not job:
                return jsonify({"error": "Invalid deploy ID"}), 404
            return jsonify(job)



from abc import ABC, abstractmethod
from typing import Any, Optional
import click
from pydantic import PrivateAttr
from visivo.models.sources.source import Source
from sqlalchemy import create_engine, event, text, inspect
from sqlalchemy.pool import NullPool
from visivo.logger.logger import Logger
import polars as pl
from copy import deepcopy
import pyarrow as pa
import json


class SqlalchemySource(Source, ABC):

    _engine: Any = PrivateAttr(default=None)
    after_connect: Optional[str] = None

    @abstractmethod
    def get_dialect(self):
        raise NotImplementedError(f"No dialect method implemented for {self.type}")

    def read_sql(self, query: str):
        with self.connect() as connection:
            query = text(query)
            results = connection.execute(query)
            columns = list(results.keys())
            data = results.fetchall()
            results.close()

        # Convert to list of dictionaries
        result_data = []
        for row in data:
            row_dict = {}
            for i, col in enumerate(columns):
                value = row[i]
                # Convert complex types to JSON strings for consistency
                if isinstance(value, (dict, list)):
                    value = json.dumps(value)
                row_dict[col] = value
            result_data.append(row_dict)

        return result_data

    def get_connection(self):

        try:
            connection = (
                self.get_engine().connect()
            )  # I wonder if creating mutltiple engines is part of the problem.
            if hasattr(self, "attach") and self.attach:
                for attachment in self.attach:
                    connection.execute(
                        text(
                            f"attach database '{attachment.source.database}' as {attachment.schema_name};"
                        )
                    )
            return connection
        except Exception as err:
            raise click.ClickException(
                f"Error connecting to source '{self.name}'. Ensure the database is running and the connection properties are correct. Full Error: {str(err)}"
            )

    def get_engine(self):

        if not self._engine:

            Logger.instance().debug(f"Creating engine for Source: {self.name}")
            self._engine = create_engine(
                self.url(), poolclass=NullPool, connect_args=self.connect_args()
            )

            @event.listens_for(self._engine, "connect")
            def connect(dbapi_connection, connection_record):
                if self.after_connect:
                    cursor_obj = dbapi_connection.cursor()
                    cursor_obj.execute(self.after_connect)
                    cursor_obj.close()

        return self._engine

    def __deepcopy__(self, memo):
        copied = self.model_copy(deep=False)
        # manually deepcopy only safe attrs
        for name, val in self.__dict__.items():
            if name != "_plugin_module":
                setattr(copied, name, deepcopy(val, memo))
        return copied

    def connect_args(self):
        return {}

    def list_databases(self):
        """Return a list of database names for this source.

        This method must be overridden by each source implementation to ensure
        proper connection testing and database discovery.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement list_databases() method. "
            "This ensures proper connection testing rather than falsely returning a default."
        )

    def introspect(self):
        """Return metadata about databases, schemas, tables, and columns efficiently."""
        result = {"name": self.name, "type": self.type, "databases": []}
        db_names = self.list_databases()

        # Use read-only engine for introspection when available (DuckDB)
        engine = None
        try:
            if hasattr(self, "get_engine") and "read_only" in self.get_engine.__code__.co_varnames:
                engine = self.get_engine(read_only=True)
            else:
                engine = self.get_engine()

            dialect = engine.dialect.name

            # Handle different database paradigms efficiently
            if dialect.startswith(("duckdb", "mysql", "snowflake")):
                # These can access multiple databases via single connection
                result["databases"] = self._introspect_via_single_connection(engine, db_names)
            elif dialect.startswith("postgresql"):
                # PostgreSQL typically needs separate connections per database
                result["databases"] = self._introspect_via_multiple_connections(db_names)
            else:
                # Default to single connection approach for other dialects
                result["databases"] = self._introspect_via_single_connection(engine, db_names)

        except Exception as e:
            Logger.instance().error(f"Error during introspection for {self.name}: {str(e)}")
            result["error"] = str(e)
        finally:
            # Clean up if we created a temporary engine
            if engine and hasattr(self, "_read_only_engine") and engine == self._read_only_engine:
                pass  # Don't dispose cached engines
            elif engine and engine != self._engine:
                engine.dispose()

        return result

    def _introspect_via_single_connection(self, engine, db_names):
        """Introspect multiple databases using a single connection."""
        databases = []
        inspector = inspect(engine)
        dialect = engine.dialect.name

        with engine.connect() as conn:
            for db_name in db_names:
                try:
                    # Switch database context if needed (MySQL/Snowflake)
                    if dialect.startswith(("mysql", "snowflake")) and db_name != self.database:
                        conn.execute(text(f"USE {db_name}"))
                        # Need new inspector after database switch
                        inspector = inspect(engine)

                    db_entry = self._introspect_database(inspector, db_name)
                    databases.append(db_entry)

                except Exception as e:
                    Logger.instance().debug(f"Failed to introspect database {db_name}: {str(e)}")
                    databases.append({"name": db_name, "error": str(e)})

        return databases

    def _introspect_via_multiple_connections(self, db_names):
        """Introspect databases that require separate connections (PostgreSQL)."""
        databases = []

        for db_name in db_names:
            engine = None
            try:
                # Only create new engine if database is different
                if db_name != self.database:
                    src_copy = deepcopy(self)
                    src_copy.database = db_name
                    src_copy._engine = None
                    if hasattr(src_copy, "_read_only_engine"):
                        src_copy._read_only_engine = None
                    engine = src_copy.get_engine()
                else:
                    engine = self.get_engine()

                inspector = inspect(engine)
                db_entry = self._introspect_database(inspector, db_name)
                databases.append(db_entry)

            except Exception as e:
                Logger.instance().debug(f"Failed to introspect database {db_name}: {str(e)}")
                databases.append({"name": db_name, "error": str(e)})
            finally:
                # Only dispose if we created a temporary engine
                if engine and db_name != self.database:
                    engine.dispose()

        return databases

    def _introspect_database(self, inspector, db_name):
        """Introspect a single database using the provided inspector."""
        try:
            schemas = inspector.get_schema_names()
        except Exception:
            schemas = []

        db_entry = {"name": db_name}

        if schemas:
            db_entry["schemas"] = []
            for schema in schemas:
                try:
                    tables = inspector.get_table_names(schema=schema)
                except Exception:
                    tables = []

                table_entries = []
                for table in tables:
                    try:
                        cols = [c["name"] for c in inspector.get_columns(table, schema=schema)]
                    except Exception:
                        cols = []
                    table_entries.append({"name": table, "columns": cols})

                db_entry["schemas"].append({"name": schema, "tables": table_entries})
        else:
            # No schemas, list tables directly
            try:
                tables = inspector.get_table_names()
            except Exception:
                tables = []

            table_entries = []
            for table in tables:
                try:
                    cols = [c["name"] for c in inspector.get_columns(table)]
                except Exception:
                    cols = []
                table_entries.append({"name": table, "columns": cols})

            db_entry["tables"] = table_entries

        return db_entry



from typing import Optional
from pydantic import BaseModel, Field
from enum import Enum


class SourceTypeEnum(str, Enum):
    sqlite = "sqlite"
    postgresql = "postgresql"
    mysql = "mysql"
    snowflake = "snowflake"
    bigquery = "bigquery"
    duckdb = "duckdb"
    excel = "xls"
    csv = "csv"


class CreateSourceRequest(BaseModel):
    project_name: str = Field(..., min_length=1)
    source_name: Optional[str] = "Example Source"
    database: Optional[str] = ""
    source_type: Optional[str] = ""
    host: Optional[str] = ""
    username: Optional[str] = ""
    password: Optional[str] = ""
    account: Optional[str] = ""
    warehouse: Optional[str] = ""
    credentials_base64: Optional[str] = ""
    project: Optional[str] = ""
    dataset: Optional[str] = ""
    project_dir: Optional[str] = ""
    port: Optional[str] = ""



import os
from flask import Flask, jsonify
from visivo.models.project import Project
from visivo.parsers.serializer import Serializer
from visivo.server.views import register_views
from visivo.logger.logger import Logger
from visivo.server.repositories.worksheet_repository import WorksheetRepository
from visivo.telemetry.middleware import init_telemetry_middleware


class FlaskApp:

    def __init__(self, output_dir, project: Project, working_dir=None):
        self.app = Flask(__name__, static_folder=output_dir, static_url_path="/data")

        self._project_json = (
            Serializer(project=project).dereference().model_dump_json(exclude_none=True)
        )
        self._project = project

        self._working_dir = working_dir
        self.hot_reload_server = None  # Will be set by serve_phase

        self.app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 0
        self.worksheet_repo = WorksheetRepository(os.path.join(output_dir, "worksheets.db"))

        # Initialize telemetry middleware
        init_telemetry_middleware(self.app, project)

        register_views(self.app, self, output_dir)

    @property
    def project(self):
        return self._project

    @project.setter
    def project(self, value):
        Logger.instance().debug(f"Setting new project on FlaskApp")
        self._project_json = (
            Serializer(project=value).dereference().model_dump_json(exclude_none=True)
        )
        self._project = value



import datetime
import json
import os
import re
from flask import jsonify, send_file, send_from_directory
from visivo.logger.logger import Logger
from visivo.utils import SCHEMA_FILE, VIEWER_PATH


def register_data_views(app, flask_app, output_dir):
    @app.route("/api/explorer/")
    def explorer_api():
        if os.path.exists(f"{output_dir}/explorer.json"):
            with open(f"{output_dir}/explorer.json", "r") as f:
                return json.load(f)
        else:
            return json.dumps({})

    @app.route("/api/schema/")
    def schema_api():
        if os.path.exists(SCHEMA_FILE):
            return send_file(SCHEMA_FILE)
        else:
            return (
                jsonify({"message": f"Schema file not found: {SCHEMA_FILE}"}),
                404,
            )

    @app.route("/api/error/")
    def error_api():
        if os.path.exists(f"{output_dir}/error.json"):
            with open(f"{output_dir}/error.json", "r") as error_file:
                return error_file.read()
        else:
            return json.dumps({})

    @app.route("/api/project/")
    def projects_api():
        return {
            "id": "id",
            "project_json": json.loads(flask_app._project_json),
            "created_at": datetime.datetime.now().isoformat(),
        }

    @app.route("/api/project_history/")
    def project_history_api():
        return [
            {
                "id": "id",
                "created_at": datetime.datetime.now().isoformat(),
            }
        ]

    @app.route("/", defaults={"path": "index.html"})
    @app.route("/<path:path>")
    def viewer_file(path):
        regex = r"\S*(\.png|\.ico|\.js|\.css|\.webmanifest|\.js\.map|\.css\.map)$"
        if re.match(regex, path):
            return send_from_directory(VIEWER_PATH, path)

        # For HTML responses, read the file and inject our scripts
        with open(os.path.join(VIEWER_PATH, "index.html"), "r") as f:
            html = f.read()

        # Add socket.io client and our hot reload script
        scripts = """
            <script src="https://cdnjs.cloudflare.com/ajax/libs/socket.io/4.0.1/socket.io.js"></script>
            <script src="/hot-reload.js"></script>
        """
        html = html.replace("</head>", f"{scripts}</head>")

        return html



"""
Command execution tracking and metrics collection.
"""

from typing import Optional, Dict, Any
from visivo.telemetry import get_telemetry_context, get_telemetry_client
from visivo.telemetry.collector import collect_project_metrics, count_filtered_jobs
from visivo.telemetry.utils import hash_project_name
from visivo.telemetry.config import is_telemetry_enabled
from visivo.telemetry.machine_id import get_machine_id
from visivo.telemetry.events import NewInstallationEvent


def track_compile_metrics(project):
    """
    Collect and track telemetry metrics during compile phase.

    Args:
        project: The project being compiled
    """
    try:
        # Check if this is the first run and send new installation event
        _check_new_installation()

        # Store hashed project name
        if project and hasattr(project, "name"):
            project_hash = hash_project_name(project.name)
            if project_hash:
                get_telemetry_context().set("project_hash", project_hash)

        # Collect object counts
        object_counts = collect_project_metrics(project)
        if object_counts:  # Only store if we successfully collected metrics
            get_telemetry_context().set("object_counts", object_counts)
    except Exception:
        # Silently ignore any telemetry errors
        pass


def track_run_metrics(project, dag_filter):
    """
    Collect and track telemetry metrics during run phase.

    Args:
        project: The project being run
        dag_filter: The DAG filter being applied
    """
    try:
        # Check if this is the first run and send new installation event
        _check_new_installation()

        # Store hashed project name
        if project and hasattr(project, "name"):
            project_hash = hash_project_name(project.name)
            if project_hash:
                get_telemetry_context().set("project_hash", project_hash)

        # Count jobs
        job_count = count_filtered_jobs(project.dag(), dag_filter)
        if job_count > 0:  # Only store if we successfully counted
            get_telemetry_context().set("job_count", job_count)
    except Exception:
        # Silently ignore any telemetry errors
        pass


def _check_new_installation():
    """
    Check if this is a new installation and send the new installation event.

    This is called from compile/run phases to ensure we catch new installations
    regardless of which command is run first.
    """
    # This will trigger the new installation event if it's a new machine ID
    # The event is sent from within get_machine_id() when creating a new ID
    get_machine_id()



"""
Event definitions and schemas for telemetry.
"""

from dataclasses import dataclass, asdict
from datetime import datetime, timezone
from typing import Dict, Any, Optional, Literal, List
import platform
import sys
import uuid
from visivo.version import VISIVO_VERSION
from .config import is_ci_environment
from .machine_id import get_machine_id


# Generate a session ID that's unique per CLI/API session but not persistent
SESSION_ID = str(uuid.uuid4())

# Get the machine ID (will be created on first use)
MACHINE_ID = None  # Lazy load to avoid I/O during import


def _get_machine_id() -> str:
    """Get machine ID, caching the result."""
    global MACHINE_ID
    if MACHINE_ID is None:
        MACHINE_ID = get_machine_id()
    return MACHINE_ID


@dataclass
class BaseEvent:
    """Base class for all telemetry events."""

    event_type: str
    timestamp: str
    session_id: str
    machine_id: str
    properties: Dict[str, Any]

    def to_dict(self) -> Dict[str, Any]:
        """Convert event to dictionary for JSON serialization."""
        return asdict(self)


@dataclass
class CLIEvent(BaseEvent):
    """Event for CLI command execution."""

    @classmethod
    def create(
        cls,
        command: str,
        command_args: List[str],
        duration_ms: int,
        success: bool,
        error_type: Optional[str] = None,
        job_count: Optional[int] = None,
        object_counts: Optional[Dict[str, int]] = None,
        project_hash: Optional[str] = None,
    ) -> "CLIEvent":
        """Create a CLI event with common properties."""
        properties = {
            "command": command,
            "command_args": command_args,
            "duration_ms": duration_ms,
            "success": success,
            "visivo_version": VISIVO_VERSION,
            "python_version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            "platform": platform.system().lower(),
            "platform_version": platform.version(),
            "architecture": platform.machine(),
            "is_ci": is_ci_environment(),
        }

        if error_type:
            properties["error_type"] = error_type

        if job_count is not None:
            properties["job_count"] = job_count

        if object_counts:
            properties["object_counts"] = object_counts

        if project_hash:
            properties["project_hash"] = project_hash

        return cls(
            event_type="cli_command",
            timestamp=datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            session_id=SESSION_ID,
            machine_id=_get_machine_id(),
            properties=properties,
        )


@dataclass
class APIEvent(BaseEvent):
    """Event for API request."""

    @classmethod
    def create(
        cls,
        endpoint: str,
        method: str,
        status_code: int,
        duration_ms: int,
        project_hash: Optional[str] = None,
    ) -> "APIEvent":
        """Create an API event with common properties."""
        properties = {
            "endpoint": endpoint,
            "method": method,
            "status_code": status_code,
            "duration_ms": duration_ms,
            "visivo_version": VISIVO_VERSION,
            "python_version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            "platform": platform.system().lower(),
            "platform_version": platform.version(),
            "architecture": platform.machine(),
            "is_ci": is_ci_environment(),
        }

        if project_hash:
            properties["project_hash"] = project_hash

        return cls(
            event_type="api_request",
            timestamp=datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            session_id=SESSION_ID,
            machine_id=_get_machine_id(),
            properties=properties,
        )


@dataclass
class NewInstallationEvent(BaseEvent):
    """Event for when a new machine ID is created (new user/installation)."""

    @classmethod
    def create(cls, machine_id: str) -> "NewInstallationEvent":
        """Create a new installation event with system properties."""
        properties = {
            "visivo_version": VISIVO_VERSION,
            "python_version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
            "platform": platform.system().lower(),
            "platform_version": platform.version(),
            "architecture": platform.machine(),
            "is_ci": is_ci_environment(),
        }

        return cls(
            event_type="new_installation",
            timestamp=datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
            session_id=SESSION_ID,
            machine_id=machine_id,  # Use the provided machine_id instead of calling _get_machine_id()
            properties=properties,
        )



"""
Machine ID management for telemetry.
"""

import os
import uuid
from pathlib import Path
from .config import is_ci_environment, is_telemetry_enabled


def get_machine_id() -> str:
    """
    Get or create a persistent anonymous machine ID.

    The machine ID is stored in ~/.visivo/machine_id and is created
    on first use. This provides anonymous user identification while
    preserving privacy.

    For CI/CD environments, generates a special prefixed ID that
    changes with each run to avoid persistence issues.

    Returns:
        str: A UUID string that uniquely identifies this installation
    """
    # Check if we're in CI/CD
    if is_ci_environment():
        # For CI, generate a new ID each time with a special prefix
        # This helps us identify CI runs and doesn't try to persist
        return f"ci-{uuid.uuid4()}"

    # Normal user environment - use persistent ID
    visivo_dir = Path.home() / ".visivo"
    machine_id_path = visivo_dir / "machine_id"

    # Try to read existing machine ID
    if machine_id_path.exists():
        try:
            with open(machine_id_path, "r") as f:
                machine_id = f.read().strip()
                # Validate it's a valid UUID (no prefix for regular users)
                if not machine_id.startswith("ci-"):
                    uuid.UUID(machine_id)
                    return machine_id
        except Exception:
            # If file is corrupted or invalid, regenerate
            pass

    # Generate new machine ID
    machine_id = str(uuid.uuid4())
    is_new_installation = True  # Flag to track this is a new installation

    # Ensure directory exists
    try:
        visivo_dir.mkdir(exist_ok=True)

        # Write machine ID to file
        with open(machine_id_path, "w") as f:
            f.write(machine_id)

    except Exception:
        # If we can't write the file, just return the generated ID
        # It will be regenerated next time, but that's better than failing
        pass

    # Send new installation event if telemetry is enabled
    # We do this after creating the machine ID to avoid circular imports
    if is_new_installation and is_telemetry_enabled():
        try:
            # Import here to avoid circular dependencies
            from .client import get_telemetry_client
            from .events import NewInstallationEvent

            client = get_telemetry_client(enabled=True)
            event = NewInstallationEvent.create(machine_id)
            client.track(event)
        except Exception:
            # Silently ignore telemetry errors
            pass

    return machine_id



from time import time

start_time = time()
from visivo.logger.logger import Logger, TypeEnum

Logger.instance().info("Starting Visivo...")
import click
import os
from dotenv import load_dotenv
from pydantic import ValidationError
import sys

from visivo.parsers.line_validation_error import LineValidationError
from visivo.telemetry import TelemetryClient, is_telemetry_enabled, get_telemetry_context
from visivo.telemetry.events import CLIEvent

from visivo.commands.dbt import dbt
from visivo.commands.deploy import deploy
from visivo.commands.serve import serve
from visivo.commands.run import run
from visivo.commands.dist import dist
from visivo.commands.compile import compile
from visivo.commands.init import init
from visivo.commands.create import create
from visivo.commands.test import test
from visivo.commands.aggregate import aggregate
from visivo.commands.archive import archive
from visivo.commands.authorize import authorize
from visivo.commands.list import list
from visivo.version import VISIVO_VERSION


@click.group()
@click.option("-p", "--profile", is_flag=True)
@click.option("-e", "--env-file", default=".env")
@click.version_option(version=VISIVO_VERSION)
def visivo(env_file, profile):
    Logger.instance().set_type(TypeEnum.spinner)
    load_env(env_file)

    # Profiling can be done with https://github.com/nschloe/tuna
    #  `tuna visivo-profile.dmp`
    # If you need to profile the import time, you can use the following command:
    #  `python -X importtime -m visivo.command_line compile 2> import.log`
    #  `tuna import.log`
    if profile:
        import cProfile
        import atexit

        Logger.instance().info("Profiling...")
        pr = cProfile.Profile()
        pr.enable()

        def exit():
            pr.disable()
            Logger.instance().info("Profiling completed")
            pr.dump_stats("visivo-profile.dmp")

        atexit.register(exit)


visivo.add_command(init)
visivo.add_command(dbt)
visivo.add_command(compile)
visivo.add_command(run)
visivo.add_command(serve)
visivo.add_command(deploy)
visivo.add_command(dist)
visivo.add_command(test)
visivo.add_command(aggregate)
visivo.add_command(archive)
visivo.add_command(authorize)
visivo.add_command(create)
visivo.add_command(list)


def load_env(env_file):
    if os.path.isfile(env_file):
        Logger.instance().debug(f"Loading env file: {env_file}")
        load_dotenv(env_file)


def print_issue_url():
    import traceback
    import urllib.parse

    stack_trace = "".join(traceback.format_exc())
    command = " ".join(sys.argv)
    issue_body = f"Command: {command}\n\nStack Trace:\n```\n{stack_trace}\n```"
    encoded_body = urllib.parse.quote(issue_body)
    issue_url = f"https://github.com/visivo-io/visivo/issues/new?body={encoded_body}"

    Logger.instance().error(
        f"\x1b]8;;{issue_url}\x1b\\Click here to report this issue\x1b]8;;\x1b\\"
    )


def _sanitize_command_args(argv):
    """
    Sanitize command arguments to remove sensitive information.

    Args:
        argv: sys.argv list

    Returns:
        tuple: (command_name, command_args)
    """
    command_name = None
    command_args = []

    if len(argv) > 1:
        command_name = argv[1]
        # Handle special cases like --version, --help
        if command_name.startswith("-"):
            command_name = "help"

        # Capture command arguments (sanitized)
        if len(argv) > 2:
            skip_next = False
            for i, arg in enumerate(argv[2:], 2):
                if skip_next:
                    command_args.append("<redacted>")
                    skip_next = False
                    continue

                # Check if this is a sensitive flag
                if arg in ["--token", "--password", "--key", "--api-key", "--secret"]:
                    command_args.append(arg)
                    skip_next = True  # Skip the next value
                # Skip file paths and values that might be sensitive
                elif arg.startswith("/") or arg.startswith("~") or "\\" in arg:
                    command_args.append("<path>")
                elif not arg.startswith("-"):
                    # This might be a value for a previous flag
                    command_args.append("<value>")
                else:
                    # Keep flags and options
                    command_args.append(arg)

    return command_name, command_args


def _track_command_execution(
    telemetry_client, command_name, command_args, execution_time, success, error_type=None
):
    """
    Track command execution telemetry.

    Args:
        telemetry_client: The telemetry client instance
        command_name: Name of the command executed
        command_args: Sanitized command arguments
        execution_time: Execution time in seconds
        success: Whether the command succeeded
        error_type: Type of error if command failed
    """
    if not telemetry_client or not command_name:
        return

    # Get any additional metrics from context
    context_data = get_telemetry_context().get_all()

    event = CLIEvent.create(
        command=command_name,
        command_args=command_args,
        duration_ms=int(execution_time * 1000),
        success=success,
        error_type=error_type,
        job_count=context_data.get("job_count") if success else None,
        object_counts=context_data.get("object_counts") if success else None,
        project_hash=context_data.get("project_hash") if success else None,
    )
    telemetry_client.track(event)


def safe_visivo():
    # Clear telemetry context for fresh start
    get_telemetry_context().clear()

    # Initialize telemetry client if enabled
    telemetry_enabled = is_telemetry_enabled()
    telemetry_client = TelemetryClient(enabled=telemetry_enabled) if telemetry_enabled else None

    # Track command execution
    command_name, command_args = _sanitize_command_args(sys.argv)
    error_type = None
    success = False

    try:

        visivo(standalone_mode=False)
        execution_time = round(time() - start_time, 2)
        Logger.instance().info(f"Visivo execution time: {execution_time}s")
        success = True

        # Track successful command
        _track_command_execution(telemetry_client, command_name, command_args, execution_time, True)

    except (ValidationError, LineValidationError) as e:
        error_type = type(e).__name__
        Logger.instance().error(str(e))
        sys.exit(1)
    except Exception as e:
        error_type = type(e).__name__
        if "STACKTRACE" in os.environ and os.environ["STACKTRACE"] == "true":
            raise e
        Logger.instance().error("An unexpected error has occurred")
        Logger.instance().error(str(e))
        Logger.instance().error(
            "To print more error information add the 'STACKTRACE=true' environment variable."
        )
        print_issue_url()
        sys.exit(1)
    finally:
        # Track failed command if an error occurred
        if not success:
            execution_time = round(time() - start_time, 2)
            _track_command_execution(
                telemetry_client, command_name, command_args, execution_time, False, error_type
            )

        # Ensure telemetry is flushed before exit
        if telemetry_client:
            telemetry_client.flush()
            telemetry_client.shutdown()


if __name__ == "__main__":
    safe_visivo()



import json
import os
import glob
import hashlib
from flask import jsonify, request, send_file

from visivo.logger.logger import Logger
from visivo.utils import get_utc_now


def register_trace_views(app, flask_app, output_dir):
    @app.route("/api/traces/<hash>/")
    def serve_trace_data_by_hash(hash):
        """API endpoint to serve trace data by hash"""
        try:
            # Find trace name by hash
            trace_dirs = glob.glob(f"{output_dir}/traces/*/")
            for trace_dir in trace_dirs:
                trace_name = os.path.basename(os.path.normpath(trace_dir))
                trace_name_hash = hashlib.md5(trace_name.encode()).hexdigest()
                if trace_name_hash == hash:
                    data_file = os.path.join(trace_dir, "data.json")
                    if os.path.exists(data_file):
                        return send_file(data_file)
                    break

            return jsonify({"message": f"Trace data not found for hash: {hash}"}), 404
        except Exception as e:
            Logger.instance().error(f"Error serving trace data by hash: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/traces/", methods=["GET"])
    def get_traces_api():
        """API endpoint for traces data"""
        try:
            # Get trace names from query parameters
            trace_names = request.args.getlist("trace_names")
            project_id = request.args.get("project_id")

            # Return traces with hash-based data URLs
            traces = []
            for name in trace_names:
                name_hash = hashlib.md5(name.encode()).hexdigest()
                traces.append(
                    {
                        "name": name,
                        "id": name,
                        "signed_data_file_url": f"/api/traces/{name_hash}/",
                    }
                )

            return jsonify(traces)
        except Exception as e:
            Logger.instance().error(f"Error fetching traces data: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/query/<project_id>/", methods=["POST"])
    def execute_query(project_id):
        try:
            data = request.get_json()
            if not data or "query" not in data:
                return jsonify({"message": "No query provided"}), 400

            query = data["query"]
            source_name = data.get("source")
            Logger.instance().info(f"Executing query with source: {source_name}")
            worksheet_id = data.get("worksheet_id")  # New: Get worksheet_id if provided

            # Get the appropriate source based on the request
            source = None
            if source_name:
                source = next(
                    (s for s in flask_app._project.sources if s.name == source_name),
                    None,
                )

            if (
                not source
                and flask_app._project.defaults
                and flask_app._project.defaults.source_name
            ):
                source = next(
                    (
                        s
                        for s in flask_app._project.sources
                        if s.name == flask_app._project.defaults.source_name
                    ),
                    None,
                )

            if not source and flask_app._project.sources:
                source = flask_app._project.sources[0]

            if not source:
                return jsonify({"message": "No source configured"}), 400

            Logger.instance().info(f"Executing query with source: {source.name}")

            # Execute the query using read_sql
            result = source.read_sql(query)

            # Transform the result into the expected format
            # result is now a list of dictionaries instead of a Polars DataFrame
            if result is None or len(result) == 0:
                response_data = {"columns": [], "rows": []}
            else:
                # Extract column names from the first row
                columns = list(result[0].keys()) if result else []
                response_data = {
                    "columns": columns,
                    "rows": result,
                }

            # If worksheet_id is provided, save the results
            if worksheet_id:
                query_stats = {
                    "timestamp": get_utc_now().isoformat(),
                    "source": source.name,
                }
                flask_app.worksheet_repo.save_results(
                    worksheet_id, json.dumps(response_data), json.dumps(query_stats)
                )

            return jsonify(response_data), 200

        except Exception as e:
            Logger.instance().error(f"Query execution error: {str(e)}")
            return jsonify({"message": str(e)}), 500

    @app.route("/api/trace/<trace_name>/query/", methods=["GET"])
    def get_trace_query(trace_name):
        try:
            query_file_path = f"{output_dir}/traces/{trace_name}/query.sql"
            if not os.path.exists(query_file_path):
                return (
                    jsonify({"message": f"Query file not found for trace: {trace_name}"}),
                    404,
                )

            with open(query_file_path, "r") as f:
                query_contents = f.read()

            return jsonify({"query": query_contents}), 200
        except Exception as e:
            Logger.instance().error(f"Error reading trace query: {str(e)}")
            return jsonify({"message": str(e)}), 500



import re
from typing import Annotated, Any, Union
from pydantic import Discriminator, Tag
from visivo.models.base.base_model import ContextStringType, RefStringType
from visivo.models.base.context_string import CONTEXT_STRING_VALUE_REGEX
from visivo.models.sources.csv_source import CSVFileSource
from visivo.models.sources.excel_source import ExcelFileSource
from visivo.models.sources.mysql_source import MysqlSource
from visivo.models.sources.postgresql_source import PostgresqlSource
from visivo.models.sources.snowflake_source import SnowflakeSource
from visivo.models.sources.sqlite_source import SqliteSource
from visivo.models.sources.bigquery_source import BigQuerySource
from visivo.models.sources.duckdb_source import DuckdbSource


def get_model_discriminator_value(value: Any) -> str:
    if isinstance(value, str) and re.search(CONTEXT_STRING_VALUE_REGEX, value):
        return "Context"
    elif isinstance(value, str):
        return "Ref"
    elif isinstance(value, dict):
        if "type" in value:
            return value["type"]
    if hasattr(value, "type"):
        return value.type

    return None


SourceField = Annotated[
    Union[
        Annotated[SqliteSource, Tag("sqlite")],
        Annotated[PostgresqlSource, Tag("postgresql")],
        Annotated[MysqlSource, Tag("mysql")],
        Annotated[SnowflakeSource, Tag("snowflake")],
        Annotated[BigQuerySource, Tag("bigquery")],
        Annotated[DuckdbSource, Tag("duckdb")],
        Annotated[CSVFileSource, Tag("csv")],
        Annotated[ExcelFileSource, Tag("xls")],
    ],
    Discriminator(get_model_discriminator_value),
]

SourceRefField = Annotated[
    Union[
        RefStringType,
        ContextStringType,
        Annotated[SqliteSource, Tag("sqlite")],
        Annotated[PostgresqlSource, Tag("postgresql")],
        Annotated[MysqlSource, Tag("mysql")],
        Annotated[SnowflakeSource, Tag("snowflake")],
        Annotated[BigQuerySource, Tag("bigquery")],
        Annotated[CSVFileSource, Tag("csv")],
        Annotated[ExcelFileSource, Tag("xls")],
    ],
    Discriminator(get_model_discriminator_value),
]



import os
import json
import base64
from collections import defaultdict
from decimal import Decimal
from datetime import datetime, date, time
from visivo.logger.logger import Logger


class Aggregator:
    @staticmethod
    def _make_json_serializable(obj):
        """Convert objects to JSON-serializable format"""
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("utf-8")
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, time):
            return obj.isoformat()
        elif isinstance(obj, list):
            return [Aggregator._make_json_serializable(item) for item in obj]
        elif isinstance(obj, dict):
            return {key: Aggregator._make_json_serializable(value) for key, value in obj.items()}
        elif (
            isinstance(obj, bool)
            or isinstance(obj, int)
            or isinstance(obj, float)
            or isinstance(obj, str)
            or obj is None
        ):
            return obj
        else:
            return str(obj)

    @classmethod
    def aggregate(cls, json_file: str, trace_dir: str):
        # Read JSON file directly with Python instead of Polars
        with open(json_file, "r") as f:
            data = json.load(f)

        # Convert column names (replace | with .)
        for row in data:
            renamed_row = {}
            for key, value in row.items():
                new_key = key.replace("|", ".")
                renamed_row[new_key] = value
            row.clear()
            row.update(renamed_row)

        cls.aggregate_data(data=data, trace_dir=trace_dir)

    @classmethod
    def aggregate_data_frame(cls, data, trace_dir: str):
        for row in data:
            renamed_row = {}
            for key, value in row.items():
                new_key = key.replace("|", ".")
                renamed_row[new_key] = value
            row.clear()
            row.update(renamed_row)

        cls.aggregate_data(data=data, trace_dir=trace_dir)

    @classmethod
    def aggregate_data(cls, data: list, trace_dir: str):
        """
        Pure Python aggregation that groups by cohort_on and aggregates other columns into lists
        """
        # Group data by cohort_on
        grouped = defaultdict(list)
        for row in data:
            cohort_on = row.get("cohort_on")
            if cohort_on is not None:
                grouped[cohort_on].append(row)

        # Aggregate each group
        result = {}
        for cohort, rows in grouped.items():
            aggregated_row = {}

            # Get all column names except cohort_on
            all_columns = set()
            for row in rows:
                all_columns.update(row.keys())
            all_columns.discard("cohort_on")

            # Aggregate each column
            for col in all_columns:
                values = []
                for row in rows:
                    if col in row:
                        values.append(row[col])

                # Only process columns that have values
                if values:
                    # If there's only one value and it is a list, unwrap it from the list
                    if len(values) == 1 and isinstance(values[0], list):
                        aggregated_row[col] = values[0]
                    else:
                        aggregated_row[col] = values

            result[cohort] = aggregated_row

        # Make result JSON-serializable (handles bytes, etc.)
        json_safe_result = cls._make_json_serializable(result)

        # Write result to JSON file
        os.makedirs(trace_dir, exist_ok=True)
        with open(f"{trace_dir}/data.json", "w") as fp:
            json.dump(json_safe_result, fp, indent=4)



import json
import os

import yaml
from visivo.logger.logger import Logger
from visivo.models.chart import Chart
from visivo.models.dashboard import Dashboard
from visivo.models.item import Item
from visivo.models.models.sql_model import SqlModel
from visivo.models.row import Row
from visivo.models.trace import Trace


def load_csv(conn, file_path, table_name):
    table_exists = (
        conn.execute(
            f"SELECT COUNT(*) FROM information_schema.tables WHERE table_name = '{table_name}'"
        ).fetchone()[0]
        > 0
    )

    if table_exists:
        Logger.instance().info(f"Table '{table_name}' already exists. Skipping creation.")
        return

    try:
        conn.execute(
            f"CREATE TABLE \"{table_name}\" AS SELECT * FROM read_csv_auto('{file_path}', encoding='UTF-8')"
        )
    except Exception:
        try:
            conn.execute(
                f"CREATE TABLE \"{table_name}\" AS SELECT * FROM read_csv_auto('{file_path}', encoding='UTF-16')"
            )
        except Exception:
            conn.execute(
                f"CREATE TABLE \"{table_name}\" AS SELECT * FROM read_csv_auto('{file_path}', ignore_errors=true)"
            )
            Logger.instance().warning(f"Loaded {os.path.basename(file_path)} with encoding errors.")


def create_example_dashboard():
    model = SqlModel(name="Example Model", sql="SELECT * FROM test_table")
    trace = Trace(
        name="Example Trace",
        model=model,
        props={"type": "scatter", "x": "?{x}", "y": "?{y}"},
    )
    chart = Chart(name="Example Chart", traces=[trace])
    return Dashboard(name="Example Dashboard", rows=[Row(items=[Item(chart=chart)])])


def write_project_file(project, project_dir):
    """
    Writes the project to a YAML file (project.visivo.yml), optionally in a given directory.
    Also writes a .gitignore file if a project_dir is provided.
    """
    project.project_file_path = (
        os.path.join(project_dir, "project.visivo.yml") if project_dir else "project.visivo.yml"
    )

    with open(project.project_file_path, "w") as f:
        content = yaml.dump(json.loads(project.model_dump_json(exclude_none=True)), sort_keys=False)
        content = content.replace("'**********'", "\"{{ env_var('DB_PASSWORD') }}\"")
        f.write(content)

    if project_dir:
        gitignore_path = os.path.join(project_dir, ".gitignore")
        lines_to_add = {".env", "target", ".visivo_cache"}

        existing_lines = set()
        if os.path.exists(gitignore_path):
            with open(gitignore_path, "r") as f:
                existing_lines = {line.strip() for line in f}

        with open(gitignore_path, "a") as f:
            for line in lines_to_add:
                if line not in existing_lines:
                    f.write(line + "\n")



from visivo.server.views.auth_views import register_auth_views
from visivo.server.views.cloud_views import register_cloud_views
from visivo.server.views.dashboard_views import register_dashboard_views
from visivo.server.views.data_views import register_data_views
from visivo.server.views.editor_views import register_editor_views
from visivo.server.views.project_views import register_project_views
from visivo.server.views.sources_views import register_source_views
from visivo.server.views.trace_views import register_trace_views
from visivo.server.views.worksheet_views import register_worksheet_views


def register_views(app, flask_app, output_dir):
    register_data_views(app, flask_app, output_dir)
    register_project_views(app, flask_app, output_dir)
    register_worksheet_views(app, flask_app, output_dir)
    register_editor_views(app, flask_app, output_dir)
    register_dashboard_views(app, flask_app, output_dir)
    register_trace_views(app, flask_app, output_dir)
    register_source_views(app, flask_app, output_dir)
    register_auth_views(app, flask_app, output_dir)
    register_cloud_views(app, flask_app, output_dir)



from typing import get_args
import click
import yaml
import json
import os
import shutil
from pathlib import Path
from git import Repo
from visivo.logger.logger import Logger
from visivo.models.include import Include
from visivo.models.models.sql_model import SqlModel
from visivo.models.sources.postgresql_source import PostgresqlSource, PostgresqlType
from visivo.models.sources.snowflake_source import SnowflakeSource, SnowflakeType
from visivo.models.sources.sqlite_source import SqliteSource, SqliteType
from visivo.models.sources.mysql_source import MysqlSource, MysqlType
from visivo.models.sources.bigquery_source import BigQuerySource, BigQueryType
from visivo.models.project import Project
from visivo.models.dashboard import Dashboard
from visivo.models.chart import Chart
from visivo.models.item import Item
from visivo.models.row import Row
from visivo.models.sources.sqlite_source import SqliteSource
from visivo.models.trace import Trace
from visivo.models.defaults import Defaults
from visivo.commands.utils import create_file_database
from visivo.parsers.file_names import PROFILE_FILE_NAME
from visivo.commands.utils import get_source_types
from visivo.models.sources.duckdb_source import DuckdbSource, DuckdbType
from visivo.parsers.parser_factory import ParserFactory
from visivo.discovery.discover import Discover
from visivo.version import VISIVO_VERSION
from visivo.models.example_type import ExampleTypeEnum


def get_env_content_for_example_type(example_type: str) -> str:
    """Generate .env content based on example type."""
    match example_type:
        case ExampleTypeEnum.github_releases:
            return "REPO_NAME=visivo\nREPO_COMPANY=visivo-io"
        case ExampleTypeEnum.ev_sales:
            return "# EV Sales Data Configuration\n# Add any required environment variables here"
        case ExampleTypeEnum.college_football:
            return (
                "# College Football 2024 Dashboard\n# Add any required environment variables here"
            )
        case _:
            return "# Add any required environment variables here"


def create_basic_project(project_name: str, project_dir: str = "."):
    """Create a basic empty project with just name and structure."""
    Logger.instance().success(f"Initializing project '{project_name}' in {project_dir}")

    # Create basic project structure
    project = Project(name=project_name)

    # Write project file
    project_file_path = os.path.join(project_dir, "project.visivo.yml")
    with open(project_file_path, "w") as fp:
        fp.write(yaml.dump(json.loads(project.model_dump_json(exclude_none=True)), sort_keys=False))

    Logger.instance().success(f"Created project file: {project_file_path}")
    return project_file_path


def load_example_project(
    project_name: str, example_type: str = ExampleTypeEnum.github_releases, project_dir: str = "."
):
    """Load example project from GitHub repository."""
    GIT_TEMP_DIR = "tempgit"

    # Map example type to actual repo name
    repo_name_map = {
        ExampleTypeEnum.github_releases: "github-releases",
        ExampleTypeEnum.ev_sales: "ev-sales",
        ExampleTypeEnum.college_football: "2024-college-football-dashboards",
    }

    repo_name = repo_name_map.get(example_type, example_type)
    repo_url = f"https://github.com/visivo-io/{repo_name}.git"
    project_path = Path(project_dir)
    env_path = project_path / ".env"
    gitignore_path = project_path / ".gitignore"
    temp_clone_path = project_path / GIT_TEMP_DIR

    Logger.instance().info(f"Loading example project '{example_type}' for '{project_name}'")

    try:
        # Backup existing .gitignore and .env
        if gitignore_path.exists():
            shutil.copy(gitignore_path, project_path / ".gitignore.bak")
        if env_path.exists():
            shutil.copy(env_path, project_path / ".env.bak")

        # Clean up any existing temp folder and clone
        if temp_clone_path.exists():
            shutil.rmtree(temp_clone_path)
        Repo.clone_from(repo_url, temp_clone_path)

        # Move files from temp into the main project path, skipping .gitignore and .env
        for item in temp_clone_path.iterdir():
            if item.name in [".git", ".gitignore", ".env", "README.md", "LICENSE"]:
                continue
            dest = project_path / item.name
            if item.is_dir():
                shutil.copytree(item, dest, dirs_exist_ok=True)
            else:
                shutil.copy2(item, dest)

        # Remove the temporary directory
        shutil.rmtree(temp_clone_path)

        # Restore .gitignore and .env if needed
        if (project_path / ".gitignore.bak").exists():
            shutil.move(project_path / ".gitignore.bak", gitignore_path)
        if (project_path / ".env.bak").exists():
            shutil.move(project_path / ".env.bak", env_path)

        # If .env doesn't exist, create it with content based on example type
        if not env_path.exists():
            env_content = get_env_content_for_example_type(example_type)
            with open(env_path, "w") as fp:
                fp.write(env_content)

        # Update project name in the YAML file
        discover = Discover(working_dir=project_path, output_dir=None)

        from ruamel.yaml import YAML

        yaml_parser = YAML()
        yaml_parser.preserve_quotes = True
        yaml_parser.indent(mapping=2, sequence=4, offset=2)

        if discover.project_file.exists():
            with discover.project_file.open("r") as f:
                data = yaml_parser.load(f)

            if "name" in data:
                data["name"] = project_name

            with discover.project_file.open("w") as f:
                yaml_parser.dump(data, f)

        Logger.instance().success(f"Successfully loaded example project '{example_type}'")
        return str(discover.project_file) if discover.project_file.exists() else None

    except Exception as e:
        Logger.instance().error(f"Error loading example project: {str(e)}")
        raise


def init_phase(project_dir, example_type=None):
    """Main init phase function that handles both simple and example project creation."""
    if project_dir:
        project_name = os.path.basename(os.path.abspath(project_dir))
    else:
        project_dir = "."
        project_name = os.path.basename(os.path.abspath("."))

    if example_type:
        # Load example project
        return load_example_project(project_name, example_type, project_dir)
    else:
        # Create basic project
        return create_basic_project(project_name, project_dir)



import os
import click
import fnmatch
import re
from pathlib import Path
from typing import List
from visivo.parsers.file_names import PROJECT_FILE_NAME
from visivo.commands.utils import get_profile_file
from visivo.utils import load_yaml_file
from visivo.logger.logger import Logger


class Discover:
    def __init__(
        self,
        working_dir: str,
        output_dir: str,
        home_dir=os.path.expanduser("~"),
    ):
        self.working_dir = working_dir
        self.home_dir = home_dir
        self.output_dir = output_dir

    @property
    def project_file(self):
        return Path(self._get_any_project_file(self.working_dir))

    @property
    def project_file_exists(self):
        return self._get_file_exists_status(self.working_dir)

    @property
    def files(self):
        if not os.path.exists(self.project_file):
            raise click.ClickException(f'Project file "{PROJECT_FILE_NAME}" not found')
        files = [self.project_file]

        self.__add_includes(files=files, file=self.project_file)

        profile_file = get_profile_file(home_dir=self.home_dir)
        if os.path.exists(profile_file):
            files.append(profile_file)

        return files

    def __add_dbt(self, data):
        from visivo.models.dbt import Dbt

        if "dbt" in data and data["dbt"]:
            dbt = Dbt(**data["dbt"])
            if dbt.enabled:
                return dbt.get_output_file(output_dir=self.output_dir, working_dir=self.working_dir)

    def __should_exclude_file(self, file_path: str, exclusions: List[str], base_path: str) -> bool:
        """Check if a file should be excluded based on exclusion patterns."""
        if not exclusions:
            return False

        # Get relative path from base_path for pattern matching
        try:
            rel_path = os.path.relpath(file_path, base_path)
        except ValueError:
            # If we can't get relative path, use the full path
            rel_path = file_path

        file_name = os.path.basename(file_path)

        for exclusion in exclusions:
            # Try different pattern matching approaches

            # 1. Direct filename match
            if exclusion == file_name:
                return True

            # 2. Glob pattern match on filename
            if fnmatch.fnmatch(file_name, exclusion):
                return True

            # 3. Glob pattern match on relative path
            if fnmatch.fnmatch(rel_path, exclusion):
                return True

            # 4. Unix-style path pattern (e.g., */config/*)
            unix_path = rel_path.replace(os.sep, "/")
            if fnmatch.fnmatch(unix_path, exclusion):
                return True

            # 5. Handle patterns like */temp/* by testing both with and without leading */
            if exclusion.startswith("*/"):
                # Try the pattern without the leading */
                pattern_without_prefix = exclusion[2:]  # Remove the */
                if fnmatch.fnmatch(unix_path, pattern_without_prefix):
                    return True
                # Also try matching any part of the path
                path_parts = unix_path.split("/")
                for i in range(len(path_parts)):
                    partial_path = "/".join(path_parts[i:])
                    if fnmatch.fnmatch(partial_path, pattern_without_prefix):
                        return True

            # 6. Try as regex pattern (catch any regex errors)
            try:
                if re.search(exclusion, rel_path) or re.search(exclusion, file_name):
                    return True
            except re.error:
                pass  # Invalid regex, skip

        return False

    def __find_yaml_files_in_directory(
        self, directory_path: str, depth: int = None, exclusions: List[str] = None
    ) -> List[str]:
        """Find YAML files in a directory with optional depth limit and exclusions."""
        yaml_files = []
        exclusions = exclusions or []

        if depth == 0:
            # Only search current directory
            for file in os.listdir(directory_path):
                file_path = os.path.join(directory_path, file)
                if os.path.isfile(file_path) and file.lower().endswith((".yml", ".yaml")):
                    if not self.__should_exclude_file(file_path, exclusions, directory_path):
                        yaml_files.append(file_path)
        else:
            # Recursive search with depth limit
            for root, dirs, files in os.walk(directory_path):
                # Calculate current depth
                current_depth = root[len(directory_path) :].count(os.sep)

                # If we have a depth limit and we've exceeded it, skip this level
                if depth is not None and current_depth >= depth:
                    dirs[:] = []  # Don't go deeper
                    continue

                # Check files in current directory
                for file in files:
                    if file.lower().endswith((".yml", ".yaml")):
                        file_path = os.path.join(root, file)
                        if not self.__should_exclude_file(file_path, exclusions, directory_path):
                            yaml_files.append(file_path)

                # Filter directories based on exclusions to avoid walking excluded paths
                dirs[:] = [
                    d
                    for d in dirs
                    if not self.__should_exclude_file(
                        os.path.join(root, d), exclusions, directory_path
                    )
                ]

        return sorted(yaml_files)  # Sort for consistent ordering

    def __add_includes(self, files, file):
        from visivo.models.include import Include

        data = load_yaml_file(file)
        base_path = os.path.dirname(file)

        output_file = self.__add_dbt(data=data)
        if output_file and os.path.exists(output_file):
            files.append(Path(output_file))

        if "includes" in data:
            for include_data in data["includes"]:
                include = Include(**include_data)
                include_path = f"{base_path}/{include.path}"

                # Handle Git repositories (existing functionality)
                if ".git" in include.path:
                    include_path = self.__get_project_file_from_git(git_url=include.path)
                    if not os.path.exists(include_path):
                        raise click.ClickException(
                            f'Invalid "include" in project. "{include_path}" referenced in "{file}" does not exist.'
                        )
                    files.append(Path(include_path))
                    self.__add_includes(files=files, file=include_path)
                    continue

                # Check if path exists
                if not os.path.exists(include_path):
                    raise click.ClickException(
                        f'Invalid "include" in project. "{include_path}" referenced in "{file}" does not exist.'
                    )

                # Handle directory inclusion
                if os.path.isdir(include_path):
                    yaml_files = self.__find_yaml_files_in_directory(
                        directory_path=include_path,
                        depth=include.depth,
                        exclusions=include.exclusions,
                    )

                    if not yaml_files:
                        # Optionally warn if no YAML files found in directory
                        click.echo(
                            f'Warning: No YAML files found in directory "{include_path}"', err=True
                        )

                    for yaml_file in yaml_files:
                        files.append(Path(yaml_file))
                        self.__add_includes(files=files, file=yaml_file)

                # Handle single file inclusion (existing functionality)
                else:
                    files.append(Path(include_path))
                    self.__add_includes(files=files, file=include_path)

    def _get_any_project_file(self, dir):
        if os.path.exists(f"{dir}/visivo_project.yml"):
            return f"{dir}/visivo_project.yml"

        return f"{dir}/{PROJECT_FILE_NAME}"

    def _get_file_exists_status(self, dir):
        if dir == "":
            return os.path.exists(f"{PROJECT_FILE_NAME}")
        return os.path.exists(f"{dir}/{PROJECT_FILE_NAME}")

    def __get_project_file_from_git(self, git_url):
        from git import Repo

        deps_folder = f"{self.working_dir}/.visivo_cache"
        if not os.path.exists(deps_folder):
            os.makedirs(deps_folder)
        if "@" not in git_url:
            raise click.ClickException(
                f'Invalid github dependency "{git_url}". A version specified with "@" is required'
            )

        repo_url = f"https://github.com/{git_url.split('@')[0]}"
        version = git_url.split("@")[1]
        if "--" in version:
            file = version.split("--")[1].strip()
            version = version.split("--")[0].strip()
        else:
            file = None

        local_folder = f"{deps_folder}/{git_url.split('@')[0].replace('.git', '')}@{version}"
        if os.path.exists(local_folder):
            repo = Repo(local_folder)
        else:
            repo = Repo.clone_from(repo_url, local_folder)

        try:
            repo.git.checkout(version)
        except Exception as e:
            repo.close()
            raise click.ClickException(f'Error cloning "{git_url}": {e}')

        if file:
            return f"{local_folder}/{file}"

        return self._get_any_project_file(local_folder)



from visivo.utils import DIST_PATH
from visivo.logger.logger import Logger
import traceback


def dist_phase(output_dir, dist_dir, deployment_root: str = None):
    import os
    import json
    import shutil
    from glob import glob
    import datetime
    from visivo.utils import get_dashboards_dir

    Logger.instance().info("Creating distribution for project in folder...")

    os.makedirs(f"{dist_dir}/data", exist_ok=True)

    deployment_root = deployment_root or ""

    try:
        dashboards_dir = get_dashboards_dir(output_dir)
        if os.path.exists(dashboards_dir):
            dist_dashboards_dir = os.path.join(dist_dir, "data", "dashboards")
            shutil.copytree(dashboards_dir, dist_dashboards_dir, dirs_exist_ok=True)

        with open(f"{output_dir}/project.json", "r") as f:
            project_json = json.load(f)
        created_at = (datetime.datetime.now().isoformat(),)
        with open(f"{dist_dir}/data/project.json", "w") as f:
            f.write(
                json.dumps(
                    {
                        "id": "id",
                        "project_json": project_json,
                        "created_at": created_at,
                    }
                )
            )
        with open(f"{dist_dir}/_redirects", "w") as f:
            f.write("/*    /index.html   200")

        with open(f"{dist_dir}/data/error.json", "w") as f:
            f.write(json.dumps({}))
        with open(f"{dist_dir}/data/project_history.json", "w") as f:
            f.write(json.dumps([{"created_at": created_at, "id": "id"}]))

        # Copy explorer.json if it exists
        if os.path.exists(f"{output_dir}/explorer.json"):
            shutil.copyfile(f"{output_dir}/explorer.json", f"{dist_dir}/data/explorer.json")

        # Generate traces.json for dist mode
        import hashlib

        trace_dirs = glob(f"{output_dir}/traces/*/", recursive=True)
        traces_list = []
        os.makedirs(f"{dist_dir}/data/traces", exist_ok=True)

        for trace_dir in trace_dirs:
            trace_name = os.path.basename(os.path.normpath(trace_dir))
            if os.path.exists(f"{output_dir}/traces/{trace_name}/data.json"):
                # Create hash-based filename for trace data
                trace_name_hash = hashlib.md5(trace_name.encode()).hexdigest()
                shutil.copyfile(
                    f"{output_dir}/traces/{trace_name}/data.json",
                    f"{dist_dir}/data/traces/{trace_name_hash}.json",
                )
                # Add trace info for traces.json
                traces_list.append(
                    {
                        "name": trace_name,
                        "id": trace_name,
                        "signed_data_file_url": f"{deployment_root}/data/traces/{trace_name_hash}.json",
                    }
                )

        # Write traces.json
        with open(f"{dist_dir}/data/traces.json", "w") as f:
            json.dump(traces_list, f)

        # Generate dashboard JSON files for dist mode
        if "dashboards" in project_json:
            os.makedirs(f"{dist_dir}/data/dashboards", exist_ok=True)
            for dashboard in project_json["dashboards"]:
                dashboard_name = dashboard["name"]
                dashboard_name_hash = hashlib.md5(dashboard_name.encode()).hexdigest()
                thumbnail_path = os.path.join(
                    dist_dir, "data", "dashboards", f"{dashboard_name_hash}.png"
                )
                thumbnail_exists = os.path.exists(thumbnail_path)

                dashboard_data = {
                    "id": dashboard_name,
                    "name": dashboard_name,
                    "signed_thumbnail_file_url": (
                        f"{deployment_root}/data/dashboards/{dashboard_name_hash}.png"
                        if thumbnail_exists
                        else None
                    ),
                }

                # Write individual dashboard JSON file using hash-based filename
                with open(f"{dist_dir}/data/dashboards/{dashboard_name_hash}.json", "w") as f:
                    json.dump(dashboard_data, f)

        shutil.copytree(DIST_PATH, dist_dir, dirs_exist_ok=True)

        index_html_path = os.path.join(dist_dir, "index.html")
        Logger.instance().info(f"Setting deployment root to {deployment_root} in index.html")
        if os.path.exists(index_html_path):
            with open(index_html_path, "r") as f:
                content = f.read()

            content = content.replace(
                "window.deploymentRoot = '';",
                f"window.deploymentRoot = '{deployment_root}';",
            )
            content = content.replace(
                'href="/',
                f'href="{deployment_root}/',
            )
            content = content.replace(
                'src="/',
                f'src="{deployment_root}/',
            )

            with open(index_html_path, "w") as f:
                f.write(content)

            site_webmanifest_path = os.path.join(dist_dir, "site.webmanifest")
            if os.path.exists(site_webmanifest_path):
                with open(site_webmanifest_path, "r") as f:
                    content = f.read()
                content = content.replace(
                    '"src": "/',
                    f'"src": "{deployment_root}/',
                )

                with open(site_webmanifest_path, "w") as f:
                    f.write(content)
    except Exception as e:
        Logger.instance().error(
            f"Error creating dist. Try running `visivo run` to ensure your project is up to date."
        )
        Logger.instance().error(f"Message: {e}, set STACKTRACE=true to see full error")
        if os.environ.get("STACKTRACE"):
            Logger.instance().error(f"{traceback.format_exc()}")
            raise e
